from __future__ import print_function
from petram.mfem_viewer import MFEM_menus
from functools import wraps
import threading

import os
import sys
import wx
import traceback
import numpy as np
import weakref
import subprocess as sp
import petram.helper.pickle_wrapper as pk
import binascii
from collections import defaultdict
from weakref import WeakKeyDictionary as WKD

import ifigure.widgets.dialog as dialog
import ifigure.events
from ifigure.utils.cbook import BuildMenu
from ifigure.utils.edit_list import EditListPanel
from ifigure.utils.edit_list import EDITLIST_CHANGED
from ifigure.utils.edit_list import EDITLIST_CHANGING
from ifigure.utils.edit_list import EDITLIST_SETFOCUS
from ifigure.widgets.miniframe_with_windowlist import DialogWithWindowList, MiniFrameWithWindowList, WithWindowList_MixIn

from petram.pi.simple_frame_plus import SimpleFramePlus

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('Dlg_plot_sol')

pane_colour1 = wx.Colour(245, 245, 245, 255)
pane_colour2 = wx.Colour(235, 235, 235, 255)


def setup_figure(fig, fig2):
    fig.nsec(1)
    fig.threed('on')
    fig.property(fig.get_axes(0), 'axis', False)
    fig.get_page(0).set_nomargin(True)
    fig.property(fig.get_page(0), 'bgcolor', 'white')
    xlim = fig2.xlim()
    ylim = fig2.ylim()
    zlim = fig2.zlim()
    fig.xlim(xlim)
    fig.ylim(ylim)
    fig.zlim(zlim)


def read_solinfo_remote(user, server, path):
    txt = "$PetraM/bin/get_soldir_info.sh " + path
    command = ["ssh",  "-o",
               "PasswordAuthentication=no",
               "-o",
               "PreferredAuthentications=publickey",
               user + '@' + server, txt]

    p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.STDOUT)

    try:
        timeout = False
        outs, errs = p.communicate(timeout=30)
    except sp.TimeoutExpired:
        timeout = True
        p.kill()
        outs, errs = p.communicate()
    res = [x.strip() for x in outs.decode('utf-8').split('\n')]
    for x in res:
        if x.find("Permission denied") != -1:
            assert False, "Connection Failed\n" + "\n".join(res)

    if timeout:
        assert False, "Connection timeout"

    res = [x for x in res if len(x) > 0]
    res = res[-1].strip()

    try:
        res2 = pk.loads(binascii.a2b_hex(res))
    except binascii.Error:
        print("Failed to call: "+" ".join(command))
        print("res is :", res)
        raise

    if not res2[0]:
        assert False, res2[1]

    return res2[1]


ThreadEnd = wx.NewEventType()
EVT_THREADEND = wx.PyEventBinder(ThreadEnd, 1)


class _XY(tuple):
    def __call__(self, value):
        return (0, 0, 1., -value)


class _YZ(tuple):
    def __call__(self, value):
        return (1, 0, 0., -value)


class _ZX(tuple):
    def __call__(self, value):
        return (0, 1., 0., -value)


def get_mapper(mesh_in):
    from petram.mesh.mesh_utils import FaceOf, EdgeOf, PointOf

    def mapper1(*args):
        return FaceOf(args, mesh=mesh_in)

    def mapper2(*args):
        return EdgeOf(args, mesh=mesh_in)

    def mapper3(*args):
        return PointOf(args, mesh=mesh_in)
    return mapper1, mapper2, mapper3


def run_in_piScope_thread(func):
    @wraps(func)
    def func2(self, *args, **kwargs):
        title = self.GetTitle()
        app = wx.GetApp().TopWindow
        petram = app.proj.setting.parameters.eval('PetraM')

        if self._plot_thread is not None:
            if self._plot_thread.is_alive():
                wx.CallAfter(dialog.showtraceback,
                             parent=self,
                             txt='Previous Job is Running',
                             title='Error',
                             traceback='')
                return
        self.SetTitle(title + '(*** processing ***)')
        maxt = app.aconfig.setting['max_thread']
        if len(app.logw.threadlist) < maxt:
            args = (self,) + args
            t = threading.Thread(target=func, args=args, kwargs=kwargs)
            self._plot_thread = t
            petram._status = 'evaluating sol...'
            ifigure.events.SendThreadStartEvent(petram,
                                                w=app,
                                                thread=t,
                                                useProcessEvent=True)
    return func2


# class DlgPlotSol(MiniFrameWithWindowList):
# class DlgPlotSol(DialogWithWindowList):


class DlgPlotSol(SimpleFramePlus):
    def __init__(self, parent, id=wx.ID_ANY, title='Plot Solution', **kwargs):
        '''
        (use this style if miniframe is used)
        style=(wx.CAPTION|
                       wx.CLOSE_BOX|
                       wx.MINIMIZE_BOX|
                       wx.RESIZE_BORDER|
                       wx.FRAME_FLOAT_ON_PARENT)
        '''
        # style =  wx.CAPTION|wx.CLOSE_BOX#|wx.RESIZE_BORDER
        style = (wx.CAPTION |
                 wx.CLOSE_BOX |
                 wx.MINIMIZE_BOX |
                 wx.RESIZE_BORDER |
                 wx.FRAME_FLOAT_ON_PARENT |
                 wx.FRAME_TOOL_WINDOW)

        from petram.sol.evaluators import def_config
        self.config = def_config

        from petram.debug import debug_evaluator_mp
        self.config["mp_debug"] = debug_evaluator_mp

        remote = parent.model.param.eval('remote')
        if remote is not None:
            host = parent.model.param.eval('host')
            self.config['cs_soldir'] = remote['rwdir']
            self.config['cs_server'] = host.getvar('server')
            self.config['cs_user'] = host.getvar('user')
            self.config['cs_ssh_opts'] = host.get_multiplex_opts()

        super(
            DlgPlotSol,
            self).__init__(
            parent,
            id,
            title,
            style=style,
            **kwargs)

        self.nb = wx.Notebook(self)
        box = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(box)
        box.Add(self.nb, 1, wx.EXPAND | wx.ALL, 1)

        tabs = ['GeomBdr', 'Points', 'Edge', 'Bdr', 'Bdr(arrow)', 'Slice',
                'Probe', 'Integral', 'Config']
        self.pages = {}
        self.elps = {}
        for t in tabs:
            p = wx.Panel(self.nb)
            p.SetBackgroundColour(pane_colour1)
            self.nb.AddPage(p, t)
            self.pages[t] = p

        self.local_soldir = None
        self.local_solsubdir = None

        # these are sol info
        self.local_sols = None
        self.remote_sols = None

        self._plot_thread = None
        self.use_profiler = False  # debug

        text = 'all'
        mfem_model = parent.model.param.getvar('mfem_model')

        if 'GeomBdr' in tabs:
            p = self.pages['GeomBdr']
            vbox = wx.BoxSizer(wx.VERTICAL)
            p.SetSizer(vbox)

            choices = list(mfem_model['Phys'])
            choices = [mfem_model['Phys'][c].fullpath() for c in choices]

            if len(choices) == 0:
                choices = ['no physcs in mode']
            ll = [['x', 'x', 0, {}],
                  ['y', 'y', 0, {}],
                  ['z', 'z', 0, {}],
                  ['Boundary Index', text, 0, {}],
                  ['NameSpace', choices[0], 4, {'style': wx.CB_READONLY,
                                                'choices': choices}],
                  ['Color', ['blue', 'none'], 506, {}],
                  [None, True, 3, {"text": 'merge solutions'}],
                  [None, True, 3, {"text": 'keep surface separated'}],
                  [None, True, 3, {"text": 'show edge only'}], ]

            elp = EditListPanel(p, ll)
            vbox.Add(elp, 1, wx.EXPAND | wx.ALL, 1)
            self.elps['GeomBdr'] = elp

            hbox = wx.BoxSizer(wx.HORIZONTAL)
            vbox.Add(hbox, 0, wx.EXPAND | wx.ALL, 5)
            ebutton = wx.Button(p, wx.ID_ANY, "Export")
            button = wx.Button(p, wx.ID_ANY, "Apply")
            ebutton.Bind(wx.EVT_BUTTON, self.onExport)
            button.Bind(wx.EVT_BUTTON, self.onApply)
            hbox.Add(ebutton, 0, wx.ALL, 1)
            hbox.AddStretchSpacer()
            hbox.Add(button, 0, wx.ALL, 1)

        if 'Points' in tabs:
            p = self.pages['Points']
            vbox = wx.BoxSizer(wx.VERTICAL)
            p.SetSizer(vbox)

            choices = list(mfem_model['Phys'])
            choices = [mfem_model['Phys'][c].fullpath() for c in choices]

            if len(choices) == 0:
                choices = ['no physcs in model']

            elp1 = [['x:', '0', 500, {}],
                    ['y:', '0', 500, {}],
                    ['z:', '0', 500, {}], ]
            tip1 = ('x coordiantes of points',
                    'y coordiantes of points',
                    'z coordiantes of points',)

            elp2 = [['start point:', '0., 0., 0.', 0, {}],
                    ['end point:', '1., 0., 0.', 0, {}],
                    ['resolution:', '30', 0, {}], ]
            tip2 = ('first edge of line',
                    'second  of line',
                    'number of points on line')

            ss = [None, None, 34, ({'text': '',
                                    'choices': ['XYZ', 'Line  '],
                                    'cb_tip': "Point could generation mode",
                                    'call_fit': False},
                                   {'elp': elp1, 'tip': tip1},
                                   {'elp': elp2, 'tip': tip2},)]
            ll = [
                [
                    'Expression', '', 0, {}], ss, [
                    'Selection', 'all', 4, {
                        'style': wx.CB_DROPDOWN, 'choices': [
                            'all', 'visible', 'hidden']}], [
                        'NameSpace', choices[0], 4, {
                            'style': wx.CB_READONLY, 'choices': choices}], [
                                None, False, 3, {
                                    "text": 'dynamic extension'}], ]

            tip = ("Expression",
                   None,
                   "Selection of domain to evaluate points",
                   "Namespace used to evaluate variabls",
                   "Generate animation using phasing")
            elp = EditListPanel(p, ll, tip=tip)
            vbox.Add(elp, 1, wx.EXPAND | wx.ALL, 1)
            self.elps['Points'] = elp

            hbox = wx.BoxSizer(wx.HORIZONTAL)
            vbox.Add(hbox, 0, wx.EXPAND | wx.ALL, 5)
            ebutton = wx.Button(p, wx.ID_ANY, "Export")
            button = wx.Button(p, wx.ID_ANY, "Apply")
            ebutton.Bind(wx.EVT_BUTTON, self.onExport)
            ebutton.Bind(wx.EVT_RIGHT_UP, self.onExportR)
            button.Bind(wx.EVT_BUTTON, self.onApply)
            hbox.Add(ebutton, 0, wx.ALL, 1)
            hbox.AddStretchSpacer()
            hbox.Add(button, 0, wx.ALL, 1)

        if 'Edge' in tabs:
            p = self.pages['Edge']
            vbox = wx.BoxSizer(wx.VERTICAL)
            p.SetSizer(vbox)

            choices = list(mfem_model['Phys'])
            choices = [mfem_model['Phys'][c].fullpath() for c in choices]

            if len(choices) == 0:
                choices = ['no physcs in model']

            s4 = {"style": wx.TE_PROCESS_ENTER,
                  "choices": [str(x + 1) for x in range(10)]}
            ll = [['Expression', '', 500, {}],
                  ["-> more...", None, None,
                      {"tlb_resize_samewidth": True, "colour": pane_colour2}, ],
                  ['Expression(x)', '', 500, {}],
                  ["<-"],
                  ['Selection', 'all', 4, {'style': wx.CB_DROPDOWN,
                                           'choices': ['all', 'visible', 'hidden']}],
                  ['NameSpace', choices[0], 4, {'style': wx.CB_READONLY,
                                                'choices': choices}],
                  [None, False, 3, {"text": 'dynamic extension'}],
                  [None, True, 3, {"text": 'merge solutions'}],
                  ['Refine', 1, 104, s4],
                  [None, True, 3, {"text": 'averaging'}], ]

            tip = ("Expression", None, "Expression for x (2D plot)",
                   None,
                   "Selection of Boundary(2D geometry) or Domain (1D geometry) to plot",
                   "Namespace to evaluate variables",
                   "Gerante animation using phasing",
                   None,
                   "Data point refinement",
                   None)
            elp = EditListPanel(p, ll, tip=tip)
            vbox.Add(elp, 1, wx.EXPAND | wx.ALL, 1)
            self.elps['Edge'] = elp

            hbox = wx.BoxSizer(wx.HORIZONTAL)
            vbox.Add(hbox, 0, wx.EXPAND | wx.ALL, 5)
            #ibutton=wx.Button(p, wx.ID_ANY, "Integrate")
            ebutton = wx.Button(p, wx.ID_ANY, "Export")
            button = wx.Button(p, wx.ID_ANY, "Apply")
            #ibutton.Bind(wx.EVT_BUTTON, self.onInteg)
            ebutton.Bind(wx.EVT_BUTTON, self.onExport)
            ebutton.Bind(wx.EVT_RIGHT_UP, self.onExportR)
            button.Bind(wx.EVT_BUTTON, self.onApply)
            hbox.Add(ebutton, 0, wx.ALL, 1)
            #hbox.Add(ibutton, 0, wx.ALL,1)
            hbox.AddStretchSpacer()
            hbox.Add(button, 0, wx.ALL, 1)
            # button.Enable(False)

        if 'Bdr' in tabs:
            p = self.pages['Bdr']
            vbox = wx.BoxSizer(wx.VERTICAL)
            p.SetSizer(vbox)

            choices = list(mfem_model['Phys'])
            choices = [mfem_model['Phys'][c].fullpath() for c in choices]

            if len(choices) == 0:
                choices = ['no physcs in model']

            s4 = {"style": wx.TE_PROCESS_ENTER,
                  "choices": [str(x + 1) for x in range(10)]}
            ll = [['Expression', '', 500, {}],
                  ["-> more...", None, None,
                      {"tlb_resize_samewidth": True, "colour": pane_colour2}, ],
                  ['Offset X', '0', 500, {}],
                  ['Offset Y', '0', 500, {}],
                  ['Offset Z', '0', 500, {}],
                  ['Scale', '1.0', 500, {}],
                  ["<-"],
                  ['Selection', 'all', 4, {'style': wx.CB_DROPDOWN,
                                           'choices': ['all', 'visible', 'hidden']}],
                  ['NameSpace', choices[0], 4, {'style': wx.CB_READONLY,
                                                'choices': choices}],
                  [None, False, 3, {"text": 'dynamic extenstion'}],
                  [None, True, 3, {"text": 'merge solutions'}],
                  ['Refine', 1, 104, s4],
                  [None, True, 3, {"text": 'averaging'}],
                  ['Decimate elements', '1', 0, {}, ], ]

            tip = ("Expression", None,
                   "Offset X (evaluated using solution namespace)",
                   "Offset Y (evaluated using solution namespace)",
                   "Offset Z (evaluated using solution namespace)",
                   "Scaling coordiantes (evaluated using local namespace)", None,
                   "Selection of Boundary(3D geometry) or Domain (2D geometry) to plot",
                   "Namespace used to evaluate variables",
                   "Generate animation using phasing",
                   None,
                   "Data point refienment", None, None,)

            elp = EditListPanel(p, ll, tip=tip)
            vbox.Add(elp, 1, wx.EXPAND | wx.ALL, 1)
            self.elps['Bdr'] = elp

            hbox = wx.BoxSizer(wx.HORIZONTAL)
            vbox.Add(hbox, 0, wx.EXPAND | wx.ALL, 5)
            ibutton = wx.Button(p, wx.ID_ANY, "Integral")
            ebutton = wx.Button(p, wx.ID_ANY, "Export")
            button = wx.Button(p, wx.ID_ANY, "Apply")
            ibutton.Bind(wx.EVT_BUTTON, self.onInteg)
            ebutton.Bind(wx.EVT_LEFT_UP, self.onExport)
            ebutton.Bind(wx.EVT_RIGHT_UP, self.onExportR)
            button.Bind(wx.EVT_BUTTON, self.onApply)
            hbox.Add(ebutton, 0, wx.ALL, 1)
            hbox.Add(ibutton, 0, wx.ALL, 1)
            hbox.AddStretchSpacer()
            hbox.Add(button, 0, wx.ALL, 1)

        if 'Bdr(arrow)' in tabs:
            p = self.pages['Bdr(arrow)']
            vbox = wx.BoxSizer(wx.VERTICAL)
            p.SetSizer(vbox)

            choices = list(mfem_model['Phys'])
            choices = [mfem_model['Phys'][c].fullpath() for c in choices]

            if len(choices) == 0:
                choices = ['no physcs in model']
            ll = [['Expression(u)', '', 500, {}],
                  ['Expression(v)', '', 500, {}],
                  ['Expression(w)', '', 500, {}],
                  ['Selection', 'all', 4, {'style': wx.CB_DROPDOWN,
                                           'choices': ['all', 'visible', 'hidden']}],
                  ['NameSpace', choices[0], 4, {'style': wx.CB_READONLY,
                                                'choices': choices}],
                  [None, False, 3, {
                      "text": 'dynamic extension (does not work)'}],
                  [None, True, 3, {"text": 'merge solutions'}],
                  ['Arrow count', 300, 400, None], ]

            tip = ("Expressoin for arrow (u)", "Expressoin for arrow (v)", "Expressoin for arrow (w)",
                   "Selection of Boundary(3D geometry) or Domain (2D geometry) to plot",
                   "Namespace used to evaluate variables",
                   None, None, "Number of arrows")
            elp = EditListPanel(p, ll, tip=tip)
            vbox.Add(elp, 1, wx.EXPAND | wx.ALL, 1)
            self.elps['Bdr(arrow)'] = elp

            hbox = wx.BoxSizer(wx.HORIZONTAL)
            vbox.Add(hbox, 0, wx.EXPAND | wx.ALL, 5)
            ebutton = wx.Button(p, wx.ID_ANY, "Export")
            button = wx.Button(p, wx.ID_ANY, "Apply")
            ebutton.Bind(wx.EVT_BUTTON, self.onExport)
            button.Bind(wx.EVT_BUTTON, self.onApply)
            hbox.Add(ebutton, 0, wx.ALL, 1)
            hbox.AddStretchSpacer()
            hbox.Add(button, 0, wx.ALL, 1)

        if 'Slice' in tabs:
            p = self.pages['Slice']
            vbox = wx.BoxSizer(wx.VERTICAL)
            p.SetSizer(vbox)

            choices = list(mfem_model['Phys'])
            choices = [mfem_model['Phys'][c].fullpath() for c in choices]

            elp1 = [['plane (a,b,c,d):', '0, 0, 1, 0', 500, {}], ]

            elp2 = [['plane (a,b,c,d):', '0, 0, 1, 0', 500, {}],
                    ['first axis:', '1., 0., 0.', 0, {}],
                    ['resolution', '0.01', 0, {}], ]

            txt1 = "\n".join(("ax + by + cz + d = 0, The parameters (a, b, c and/or d) can be 1D array",
                              "with the same length to define multiple plane."))
            ss = [None, None, 34, ({'text': 'Rendering',
                                    'choices': ['Triangles', 'Grid'],
                                    'call_fit': False},
                                   {'elp': elp1, 'tip': [txt1]},
                                   {'elp': elp2, 'tip': [txt1,
                                                         "1st axis on the plane, 2nd one is normal to the first axis and (a, b, c)",
                                                         "resolution of image"]},)]

            if len(choices) == 0:
                choices = ['no physcs in model']
            ll = [['Expression', '', 500, {}],
                  ss,
                  ['Selection', 'all', 4, {'style': wx.CB_DROPDOWN,
                                           'choices': ['all', 'visible', 'hidden']}],
                  ['NameSpace', choices[0], 4, {'style': wx.CB_READONLY,
                                                'choices': choices}],
                  [None, False, 3, {"text": 'dynamic extension'}],
                  [None, True, 3, {"text": 'merge solutions'}], ]
            tip = ("Expression",  None,
                   "Selection of volumes (3D geometry) or faces (2D)",
                   "Namespace used to evaluate variables",
                   None, None)
            elp = EditListPanel(p, ll, tip=tip)
            vbox.Add(elp, 1, wx.EXPAND | wx.ALL, 1)
            self.elps['Slice'] = elp

            hbox = wx.BoxSizer(wx.HORIZONTAL)
            vbox.Add(hbox, 0, wx.EXPAND | wx.ALL, 5)
            ebutton = wx.Button(p, wx.ID_ANY, "Export")
            button = wx.Button(p, wx.ID_ANY, "Apply")
            ebutton.Bind(wx.EVT_BUTTON, self.onExport)
            ebutton.Bind(wx.EVT_RIGHT_UP, self.onExportR)
            button.Bind(wx.EVT_BUTTON, self.onApply)
            hbox.Add(ebutton, 0, wx.ALL, 1)
            hbox.AddStretchSpacer()
            hbox.Add(button, 0, wx.ALL, 1)

        if 'Integral' in tabs:
            p = self.pages['Integral']
            vbox = wx.BoxSizer(wx.VERTICAL)
            p.SetSizer(vbox)

            choices = list(mfem_model['Phys'])
            choices = [mfem_model['Phys'][c].fullpath() for c in choices]
            if len(choices) == 0:
                choices = ['no physics in model']

            dom_bdr = ['Domain', 'Boundary']
            ll = [['Expression', '', 500, {}],
                  ['Kind', dom_bdr[0], 4, {'style': wx.CB_READONLY,
                                           'choices': dom_bdr}],
                  ['Index', text, 500, {}],
                  ['Order', '2',  0, {}],
                  ['NameSpace', choices[0], 4, {'style': wx.CB_READONLY,
                                                'choices': choices}], ]

            elp = EditListPanel(p, ll)
            vbox.Add(elp, 1, wx.EXPAND | wx.ALL, 1)
            self.elps['Integral'] = elp

            hbox = wx.BoxSizer(wx.HORIZONTAL)
            vbox.Add(hbox, 0, wx.EXPAND | wx.ALL, 5)
            button = wx.Button(p, wx.ID_ANY, "Apply")
            button.Bind(wx.EVT_BUTTON, self.onApply)
            hbox.AddStretchSpacer()
            hbox.Add(button, 0, wx.ALL, 1)

        if 'Probe' in tabs:
            p = self.pages['Probe']
            vbox = wx.BoxSizer(wx.VERTICAL)
            p.SetSizer(vbox)

            choices = list(mfem_model['Phys'])
            choices = [mfem_model['Phys'][c].fullpath() for c in choices]
            if len(choices) == 0:
                choices = ['no physics in model']

            ll = [['Expression', '', 0, {}],
                  ['Expression(x)', '', 0, {}],
                  ['NameSpace', choices[0], 4, {'style': wx.CB_READONLY,
                                                'choices': choices}], ]

            elp = EditListPanel(p, ll)
            vbox.Add(elp, 1, wx.EXPAND | wx.ALL, 1)
            self.elps['Probe'] = elp

            hbox = wx.BoxSizer(wx.HORIZONTAL)
            vbox.Add(hbox, 0, wx.EXPAND | wx.ALL, 5)
            ebutton = wx.Button(p, wx.ID_ANY, "Export")
            button = wx.Button(p, wx.ID_ANY, "Apply")
            ebutton.Bind(wx.EVT_BUTTON, self.onExport)
            button.Bind(wx.EVT_BUTTON, self.onApply)
            hbox.Add(ebutton, 0, wx.ALL, 1)
            hbox.AddStretchSpacer()
            hbox.Add(button, 0, wx.ALL, 1)

        if 'Config' in tabs:
            p = self.pages['Config']
            vbox = wx.BoxSizer(wx.VERTICAL)
            p.SetSizer(vbox)

            elp1 = [["Sol", "sol", 504, {"choices": ["sol", ],
                                         "choices_cb": self.local_sollist}],
                    ["Sub dir.", "None", 4, {"style": wx.CB_READONLY,
                                             "choices": ["", ]}, ],
                    [None, None, 141, {"alignright": True,
                                       "func": self.OnLoadLocalSol,
                                       "noexpand": True,
                                       "label": "Reload choices"}], ]
            tip1 = ("Solution folder",
                    "Subdirectory (for parametric scan/time-dependent sims.)",
                    None)
            elp2 = [["Number of workers", self.config['mp_worker'], 400, ],
                    ["Sol", "sol", 504, {"choices_cb": self.local_sollist,
                                         "choices": ["sol", ], }],
                    #                                       "UpdateUI": self.OnUpdateUI_local}],
                    ["Sub dir.", "None", 4, {"style": wx.CB_READONLY,
                                             "choices": ["", ]}, ],
                    [None, None, 141, {"alignright": True,
                                       "func": self.OnLoadLocalSol,
                                       "noexpand": True,
                                       "label": "Reload choices"}], ]
            tip2 = ("Numboer of worker processes",
                    "Solution folder",
                    "Subdirectory (for parametric scan/time-dependent sims.)",
                    None,)

            elp3 = [["Server", self.config['cs_server'], 0, ],
                    ["Number of workers", self.config['cs_worker'], 400, ],
                    ["Sol dir.", self.config['cs_soldir'], 504,
                     {"choices": [self.config['cs_soldir'], ],
                      "choices_cb": self.remote_sollist, }],
                    ["Sub dir.", "None", 4, {"style": wx.CB_READONLY,
                                             "choices": ["", ]}, ],
                    [None, None, 141, {"alignright": True,
                                       "func": self.OnLoadRemoteSol,
                                       "noexpand": True,
                                       "label": "Reload choices"}],]

            tip3 = ("Remote server name",
                    "Numboer of worker processes on remoter server",
                    "Solution directory",
                    "Subdirectory",
                    None,)

            choices = ['Single', 'MP', 'C/S']
            tip = '\n'.join(("Single: plot local solution, MP: plot local solution with multiprocessing,",
                             "C/S: plot solution on a remote server"))
            ll = [[None, None, 34, ({'text': "Worker Mode",
                                     'choices': choices,
                                     'cb_tip': tip,
                                     'call_fit': False},
                                    {'elp': elp1, 'tip': tip1},
                                    {'elp': elp2, 'tip': tip2},
                                    {'elp': elp3, 'tip': tip3},), ], ]

            elp = EditListPanel(p, ll)
            vbox.Add(elp, 1, wx.EXPAND | wx.ALL, 1)
            self.elps['Config'] = elp

            if self.config['use_cs']:
                c = choices[2]
            elif self.config['use_mp']:
                c = choices[1]
            else:
                c = choices[0]

            elp.SetValue([[c,
                           ['', 'sol', "", None],
                           [2, 'sol', "", None, ],
                           [self.config['cs_server'],
                               self.config['cs_worker'],
                               self.config['cs_soldir'],
                               '',
                               None,
                            ],
                           ]])
            parent.model.variables.setvar('remote_soldir',
                                          self.config['cs_soldir'])

        self.nb.SetSelection(self.nb.GetPageCount() - 1)
        self.Show()
        self.Layout()
        self.SetSize((850, 500))
        self.Bind(EDITLIST_CHANGED, self.onEL_Changed)
        self.Bind(EDITLIST_CHANGING, self.onEL_Changing)
        self.Bind(EDITLIST_SETFOCUS, self.onEL_SetFocus)
        self.Bind(EVT_THREADEND, self.onThreadEnd)
        wx.CallAfter(self.update_sollist_local1)
        wx.CallAfter(self.update_sollist_local2)
        wx.CallAfter(self.CentreOnParent)

        self.solvars = WKD()
        self.evaluators = {}
        self.solfiles = {}
        self.Bind(wx.EVT_CHILD_FOCUS, self.OnChildFocus)

    def name(self):
        return 'dlg_plot_sol'

    def onClose(self, evt):
        super(DlgPlotSol, self).onClose(evt)
        self.clean_evaluators()

    def clean_evaluators(self):
        for k in self.evaluators:
            self.evaluators[k].terminate_allnow()
        wx.Sleep(1)
        from petram.sol.evaluator_cs import EvaluatorClient
        for k in self.evaluators:
            if isinstance(self.evaluators[k],
                          EvaluatorClient):
                if (self.evaluators[k].p is not None and
                        self.evaluators[k].p.poll() is None):
                    self.evaluators[k].p.terminate()
        self.evaluators = {}

    def get_remote_subdir_cb(self):
        return self.elps['Config'].widgets[0][0].elps[2].widgets[3][0]

    def get_local_single_subdir_cb(self):
        return self.elps['Config'].widgets[0][0].elps[0].widgets[1][0]

    def get_local_multi_subdir_cb(self):
        return self.elps['Config'].widgets[0][0].elps[1].widgets[2][0]

    def update_subdir_local(self, path, ss1):
        single_cb2 = self.get_local_single_subdir_cb()
        multi_cb2 = self.get_local_multi_subdir_cb()

        from petram.sol.listsoldir import gather_soldirinfo
        info = gather_soldirinfo(path)

        dirnames = [""]
        choices = [""]

        solvers = list(info["checkpoint"])
        for solver in solvers:
            kk = sorted(list(info["checkpoint"][solver]))
            for k in kk:
                dirnames.append(info["checkpoint"][solver][k])
                choices.append(solver + "(" + str(k[1]) + ")")
        choices = choices + info["cases"]
        dirnames = dirnames + info["cases"]

        single_cb2.SetChoices(choices)
        multi_cb2.SetChoices(choices)

        if ss1 in dirnames:
            single_cb2.SetSelection(dirnames.index(ss1))
            multi_cb2.SetSelection(dirnames.index(ss1))
        else:
            ss1 = dirnames[0]

        probes = info["probes"]  # mapping from probe name to file
        self.local_sols = (path, probes, dict(zip(choices, dirnames)))
        return ss1

    def update_sollist_local_common(self, idx):
        model = self.GetParent().model

        sol_names = [name for name, child in model.solutions.get_children()]
        sols = [child for name, child in model.solutions.get_children()]
        owndirs = [x.owndir() for x in sols]

        single_cb1 = self.elps['Config'].widgets[0][0].elps[0].widgets[0][0]
        multi_cb1 = self.elps['Config'].widgets[0][0].elps[1].widgets[1][0]

        choices = ([single_cb1.GetString(n) for n in range(single_cb1.GetCount())] +
                   [multi_cb1.GetString(n) for n in range(multi_cb1.GetCount())])
        choices = list(set(choices))

        s1 = str(single_cb1.GetValue())
        s2 = str(multi_cb1.GetValue())

        if s1 in sol_names:
            owndir1 = owndirs[sol_names.index(s1)]
        else:
            owndir1 = s1
        if s2 in sol_names:
            owndir2 = owndirs[sol_names.index(s2)]
        else:
            owndir2 = s2

        if not s1 in sol_names:
            sol_names.append(s1)
        if not s2 in sol_names:
            sol_names.append(s2)
        for x in choices:
            if not x in sol_names:
                sol_names.append(x)

        sol_names = [x for x in sol_names if len(x) > 0]
        single_cb1.SetChoices(sol_names)
        multi_cb1.SetChoices(sol_names)

        if self.local_soldir is not None:
            ss1 = self.local_solsubdir
        else:
            if model.param.eval('sol') is not None:
                ss1 = ""
            else:
                ss1 = None

        if idx == 1:
            if os.path.exists(owndir1):
                self.update_subdir_local(owndir1, ss1)
        else:
            if os.path.exists(owndir2):
                self.update_subdir_local(owndir2, ss1)

    def update_sollist_local1(self):
        self.update_sollist_local_common(1)

    def update_sollist_local2(self):
        self.update_sollist_local_common(2)

    def update_subdir_remote(self):
        from ifigure.widgets.dialog import progressbar

        dlg = progressbar(self, 'Checking remote work directory...',
                          'In progress', 5)
        dlg.Show()
        wx.GetApp().Yield()
        try:
            info = read_solinfo_remote(self.config['cs_user'],
                                       self.config['cs_server'],
                                       self.config['cs_soldir'])

        except AssertionError as err:
            print(err.args[0])
            dlg.Destroy()
            wx.CallAfter(dialog.showtraceback, parent=self,
                         txt='Faled to read remote directory info',
                         title='Error',
                         traceback=err.args[0])
            return ""
        except:
            #_, _, tb = sys.exc_info()
            # traceback.print_tb(tb) # Fixed format
            #tb_info = traceback.extract_tb(tb)
            #filename, line, func, text = tb_info[-1]
            dlg.Destroy()
            wx.CallAfter(dialog.showtraceback, parent=self,
                         txt='Faled to read remote directory info',
                         title='Error',
                         traceback=traceback.format_exc(limit=-1))
            return ""
        dlg.Destroy()
        dirnames = [""]
        choices = [""]
        solvers = list(info["checkpoint"])
        for solver in solvers:
            kk = sorted(list(info["checkpoint"][solver]))
            for k in kk:
                dirnames.append(info["checkpoint"][solver][k])
                choices.append(solver + "(" + str(k[1]) + ")")
        choices = choices + info["cases"]
        dirnames = dirnames + info["cases"]

        cb2 = self.get_remote_subdir_cb()
        cb2.SetChoices(choices)
        ss1 = str(cb2.GetValue())
        if ss1 in choices:
            cb2.SetSelection(choices.index(ss1))

        ss1 = str(cb2.GetValue())
        probes = info["probes"]  # mapping from probe name to file
        self.remote_sols = (self.config['cs_soldir'],
                            probes, dict(zip(choices, dirnames)))
        return ss1

    def get_current_choices(self):
        if self.config['use_cs']:
            base = self.remote_sols[0]
            v = self.remote_sols[2].values()
            remote = True
        else:
            base = self.local_sols[0]
            v = self.local_sols[2].values()
            remote = False

        from string import digits

        def extract_trailing_digits(txt):
            return txt[len(txt.rstrip(digits)):]

        sorted_subs = [x[1] for x in sorted([(int(extract_trailing_digits(x)), x)
                                             for x in v if len(extract_trailing_digits(x)) != 0])]
        if '' in v:
            sorted_subs = [''] + sorted_subs

        return remote, base, sorted_subs

    def OnLoadLocalSol(self, evt):
        self.update_sollist_local1()
        self.update_sollist_local2()
        self.load_sol_if_needed()

    def OnLoadRemoteSol(self, evt):
        self.update_subdir_remote()

    def OnUpdateUI_local(self, evt):
        pass

    def OnUpdateUI_remote(self, evt):
        pass

    def local_sollist(self):
        model = self.GetParent().model
        sol_names = [name for name, child in model.solutions.get_children()]

        single_cb1 = self.elps['Config'].widgets[0][0].elps[0].widgets[0][0]
        multi_cb1 = self.elps['Config'].widgets[0][0].elps[1].widgets[1][0]

        choices = ([single_cb1.GetString(n)
                    for n in range(single_cb1.GetCount())] +
                   [multi_cb1.GetString(n)
                    for n in range(multi_cb1.GetCount())])
        choices = list(set(choices))

        s1 = str(single_cb1.GetValue())
        s2 = str(multi_cb1.GetValue())
        if not s1 in sol_names:
            sol_names.append(s1)
        if not s2 in sol_names:
            sol_names.append(s2)
        for x in choices:
            if not x in sol_names:
                sol_names.append(x)
        sol_names = [x for x in sol_names if len(x) > 0]
        return sol_names

    def remote_sollist(self):
        remote_cb1 = self.elps['Config'].widgets[0][0].elps[2].widgets[2][0]
        choices = [remote_cb1.GetString(n)
                   for n in range(remote_cb1.GetCount())]
        s1 = str(remote_cb1.GetValue())
        if not s1 in choices:
            choices.append(s1)
        choices = [x for x in choices if len(x) > 0]
        return choices

    def OnChildFocus(self, evt):
        self.GetParent()._palette_focus = 'plot'
        evt.Skip()

    def post_threadend(self, func, *args, **kwargs):
        evt = wx.PyCommandEvent(ThreadEnd, wx.ID_ANY)
        evt.pp_method = (func, args, kwargs)
        wx.PostEvent(self, evt)

    def set_title_no_status(self):
        title = self.GetTitle()
        self.SetTitle(title.split('(')[0])

    def onThreadEnd(self, evt):
        self.set_title_no_status()
        m = evt.pp_method[0]
        args = evt.pp_method[1]
        kargs = evt.pp_method[2]
        m(*args, **kargs)
        evt.Skip()

    def load_sol_if_needed(self):
        from petram.sol.solsets import read_sol, find_solfiles
        model = self.GetParent().model
        solfiles = model.variables.getvar('solfiles')

        doit = False
        if solfiles is not None:
            cpath = os.path.dirname(solfiles.set[0][0][0])
            if self.local_soldir is None:
                sol = model.param.eval('sol')
                if sol is None:
                    if model.variables.hasvar('solfiles'):
                        model.variables.delvar('solfiles')
                    return
                npath = sol.owndir()
                self.local_soldir = npath
                self.local_solsubdir = ""
            else:
                npath = os.path.join(self.local_soldir, self.local_solsubdir)
                if not os.path.exists(npath):  # fall back
                    npath = sol.owndir()
                    self.local_soldir = npath
                    self.local_solsubdir = ""

            if os.path.normpath(npath) != os.path.normpath(cpath):
                doit = True
            else:
                mfem_model = model.param.getvar('mfem_model')
                mfem_model.local_sol_path = npath
        else:
            doit = True
            if self.local_soldir is not None:
                npath = os.path.join(self.local_soldir, self.local_solsubdir)
                if not os.path.exists(npath):  # fall back
                    sol = model.param.eval('sol')
                    npath = sol.owndir()
                    self.local_soldir = npath
                    self.local_solsubdir = ""
            else:
                sol = model.param.eval('sol')
                if sol is None:
                    if model.variables.hasvar('solfiles'):
                        model.variables.delvar('solfiles')
                    return
                npath = sol.owndir()
                self.local_soldir = npath
                self.local_solsubdir = ""
        if doit:
            try:
                print("reading sol from ", npath)
                solfiles = find_solfiles(path=npath)
                if solfiles is None:
                    if model.variables.hasvar('solfiles'):
                        model.variables.delvar('solfiles')
                else:
                    model.variables.setvar('solfiles', solfiles)
                mfem_model = model.param.getvar('mfem_model')
                mfem_model.local_sol_path = npath
            except BaseException:
                traceback.print_exc()
                if model.variables.hasvar('solfiles'):
                    model.variables.delvar('solfiles')

    def onEL_Changed(self, evt):
        sel = self.nb.GetSelection()
        if sel != self.nb.GetPageCount() - 1:
            evt.Skip()
            return

        model = self.GetParent().model
        v = self.elps['Config'].GetValue()

        if str(v[0][0]) == 'Single':
            if (self.config['use_mp'] or
                    self.config['use_cs']):
                self.clean_evaluators()

            self.config['use_mp'] = False
            self.config['use_cs'] = False
            model.variables.setvar('remote_soldir', None)

            #info (path, probes, dirnames)
            sol = model.solutions.get_child(name=str(v[0][1][0]))
            if sol is None:
                tmp = os.path.expanduser(str(v[0][1][0]))
                if os.path.exists(tmp):
                    owndir = tmp
                else:
                    assert False, "Does not exits " + str(v[0][1][0])
            else:
                owndir = sol.owndir()
            if self.local_sols is None:
                self.update_sollist_local1()

            ss1 = self.local_sols[2][str(v[0][1][1])]
            ss1 = self.update_subdir_local(owndir, ss1)
            self.local_soldir = owndir
            self.local_solsubdir = ss1

            self.load_sol_if_needed()

        elif str(v[0][0]) == 'MP':
            if not self.config['use_mp']:
                self.clean_evaluators()

            if self.config['mp_worker'] != v[0][2][0]:
                self.clean_evaluators()

            self.config['mp_worker'] = v[0][2][0]
            self.config['use_mp'] = True
            self.config['use_cs'] = False

            model.variables.setvar('remote_soldir', None)

            sol = model.solutions.get_child(name=str(v[0][2][1]))
            if sol is None:
                tmp = os.path.expanduser(str(v[0][2][1]))
                if os.path.exists(tmp):
                    owndir = tmp
                else:
                    assert False, "Does not exits " + str(v[0][2][1])
            else:
                owndir = sol.owndir()

            if self.local_sols is None:
                self.update_sollist_local2()

            ss1 = self.local_sols[2][str(v[0][2][2])]
            ss1 = self.update_subdir_local(owndir, ss1)
            self.local_soldir = owndir
            self.local_solsubdir = ss1
            self.load_sol_if_needed()

        elif str(v[0][0]) == 'C/S':
            if not self.config['use_cs']:
                self.clean_evaluators()

            if self.config['cs_worker'] != v[0][3][1]:
                self.clean_evaluators()

            self.config['cs_worker'] = str(v[0][3][1])

            reload_remote = False
            if (not self.config['use_cs'] or
                self.config['cs_server'] != str(v[0][3][0]) or
                    self.config['cs_soldir'] != str(v[0][3][2])):
                reload_remote = True

            self.config['cs_server'] = str(v[0][3][0])
            self.config['cs_soldir'] = str(v[0][3][2])
            self.config['use_mp'] = False
            self.config['use_cs'] = True

            model.variables.setvar('remote_soldir', self.config['cs_soldir'])

            if reload_remote:
                self.update_subdir_remote()

            cb2 = self.get_remote_subdir_cb()
            ss1 = str(cb2.GetValue())
            # if ss1 != "":
            if self.remote_sols is not None:
                self.config['cs_solsubdir'] = str(self.remote_sols[2][ss1])
        #print('EL changed', self.config)

    def onEL_Changing(self, evt):
        pass

    def onEL_SetFocus(self, evt):
        pass

    def onApply(self, evt):
        elp = self.get_selected_elp()
        elp.AddCurrentToHistory()

        t = self.get_selected_plotmode()
        m = getattr(self, 'onApply' + t)

        m(evt)

    def onInteg(self, evt):
        elp = self.get_selected_elp()
        elp.AddCurrentToHistory()

        t = self.get_selected_plotmode()
        m = getattr(self, 'onInteg' + t)
        m(evt)

    def onExport(self, evt):
        elp = self.get_selected_elp()
        elp.AddCurrentToHistory()

        t = self.get_selected_plotmode()
        m = getattr(self, 'onExport' + t)

        m(evt)
    '''
    def onExport2(self, evt):
        t = self.get_selected_plotmode()
        m = getattr(self, 'onExport2' + t)
        m(evt)
    '''

    def onExportR(self, evt):
        elp = self.get_selected_elp()
        elp.AddCurrentToHistory()

        t = self.get_selected_plotmode()
        m1 = getattr(self, 'onExportR1' + t)
        m2 = getattr(self, 'onExportR2' + t)
        menu = wx.Menu()
        f1 = menu.Append(
            wx.ID_ANY,
            'All Subdirectories',
            'loop over subdirectoris')
        self.Bind(wx.EVT_MENU, m1, f1)
        f2 = menu.Append(wx.ID_ANY, 'Expand exp(-jwt)', '')
        self.Bind(wx.EVT_MENU, m2, f2)
        evt.GetEventObject().PopupMenu(menu, evt.GetPosition())
        menu.Destroy()
        evt.Skip()

    def get_selected_elp(self):
        t = self.nb.GetPageText(self.nb.GetSelection())
        elp = self.elps[t]
        return elp

    def get_selected_plotmode(self, kind=False):
        t = self.nb.GetPageText(self.nb.GetSelection())
        t = t.replace('(', '').replace(')', '')
        if kind:
            kinds = {'Bdr': 'bdry',
                     'BdrArrorw': 'bdry',
                     'Edge': 'edge',
                     'Slice': 'domain',
                     'Points': 'domain',
                     'Integral': 'domain/boundary',
                     'Domain': 'domain'}
            i = getattr(self, 'get_attrs_field_' + t)
            value = self.elps[t].GetValue()
            attrs = str(value[i()])
            if attrs.strip().lower() != 'all':
                attrs = [int(x) for x in attrs.split(',') if x.strip() != '']
            if t == 'Integral':
                value = self.elps['Integral'].GetValue()
                kind = str(value[1]).strip()
                return kinds[t].lower(), attrs
            else:
                return kinds[t], attrs
        else:
            return t

    def add_selection(self, sel):
        t = self.get_selected_plotmode()
        i = getattr(self, 'get_attrs_field_' + t)
        attrs = self.elps[t].GetValue()[i()]
        if attrs.strip().lower() == 'all':
            return
        attrs = sorted(set([int(x) for x in attrs.split(',')] + sel))
        self.set_selection(attrs)

    def rm_selection(self, sel):
        t = self.get_selected_plotmode()
        i = getattr(self, 'get_attrs_field_' + t)
        attrs = self.elps[t].GetValue()[i()]
        if attrs.strip().lower() == 'all':
            return
        #print(attrs, sel)
        attrs = sorted([int(x) for x in attrs.split(',') if not int(x) in sel])
        self.set_selection(attrs)

    def set_selection(self, sel):
        t = self.get_selected_plotmode()
        i = getattr(self, 'get_attrs_field_' + t)
        txt = ', '.join([str(s) for s in sel])
        v = self.elps[t].GetValue()
        v[i()] = txt
        self.elps[t].SetValue(v)

    #
    #   Edge value ('Edge' tab)
    #
    @run_in_piScope_thread
    def onApplyEdge(self, evt):
        value = self.elps['Edge'].GetValue()
        expr = str(value[0]).strip()
        expr_x = str(value[1]).strip()

        if value[4]:
            from ifigure.widgets.wave_viewer import WaveViewer
            cls = WaveViewer
        else:
            cls = None
        refine = int(value[6])

        data, data_x, battrs = self.eval_edge(mode='plot', refine=refine)
        if data is None:
            return

        self.post_threadend(self.make_plot_edge, data, battrs,
                            data_x=data_x,
                            cls=cls, expr=expr, expr_x=expr_x,
                            force_float=(not value[4]))

    # @run_in_piScope_thread
    def onExportR1Edge(self, evt):
        remote, base, subs = self.get_current_choices()
        value = self.elps['Edge'] .GetValue()
        refine = int(value[6])

        all_data = []
        for s in subs:
            if s.strip() == '':
                contineu
            if remote:
                self.config['cs_soldir'] = base
                self.config['cs_solsubdir'] = s
            else:
                self.local_soldir = base
                self.local_solsubdir = s
                self.load_sol_if_needed()

            data, data_x, battrs = self.eval_edge(mode='integ', refine=refine)
            if data is None:
                pass
            else:
                ndim = data[0][0].shape[1]
                verts = np.hstack([v.flatten() for v, c, a in data]).flatten()
                cdata = np.hstack([c.flatten() for v, c, a in data]).flatten()
                verts = verts.reshape(-1, ndim)
                data = {'vertices': verts, 'data': cdata}

                if data_x is not None:
                    cxdata = np.hstack([c.flatten()
                                        for v, c, a in data_x]).flatten()
                    xverts = np.hstack([v.flatten()
                                        for v, c, a in data_x]).flatten()
                    data['xvertices'] = xverts
                    data['xdata'] = cxdata

            all_data.append({"subdirs": s, "data": data})

        self.post_threadend(self.export_to_piScope_shell,
                            all_data, 'edge_data')

    def onExportR2Edge(self, evt):
        wx.CallAfter(
            dialog.showtraceback,
            parent=self,
            txt='Not Yet Implemented',
            title='Error',
            traceback='Exporing all time slice for frequency \ndomain analysis is not available')
        wx.CallAfter(self.set_title_no_status)

    def make_plot_edge(self, data, battrs,
                       data_x=None, cls=None,
                       expr='', expr_x='', force_float=False):
        from ifigure.interactive import figure

        if data_x is None:
            # if verts is 1D, treat it 2D plot even if data_x is None
            if data[0][0].shape[1] == 1:
                data_x = [(None, verts[:, 0]) for verts, cdata, adata in data]
                data = [(None, cdata) for verts, cdata, adata in data]

        if data_x is None:
            v = figure(viewer=cls)
            v.update(False)
            v.title(expr + ':' + str(battrs))
            setup_figure(v, self.GetParent())
            for verts, cdata, adata in data:
                if cls is None:
                    v.solid(verts, adata, cz=True, cdata=cdata.astype(float),
                            shade='linear')
                else:
                    v.solid(verts, adata, cz=True, cdata=cdata,
                            shade='linear')
            v.update(True)
            v.view('noclip')
            v.view('equal')
            v.update(False)
            ax = self.GetParent().get_axes()
            param = ax.get_axes3d_viewparam(ax._artists[0])
            ax2 = v.get_axes()
            ax2.set_axes3d_viewparam(param, ax2._artists[0])
            v.lighting(light=0.5)
            v.update(True)
        else:  # make 2D plot
            v = figure(viewer=cls)
            v.title(expr)
            v.xtitle(expr_x)
            for yy, xx in zip(data, data_x):
                y = yy[1].flatten()
                x = xx[1].flatten()
                xidx = np.argsort(x)
                if force_float:
                    yy = y[xidx]
                    if np.iscomplexobj(yy):
                        v.plot(x[xidx], yy.real)
                    else:
                        v.plot(x[xidx], yy)
                else:
                    if cls is None:
                        v.plot(x[xidx], y[xidx])
                    else:
                        data = y[xidx].astype(complex, copy=False)
                        v.plot(x[xidx], data)

    def onExportEdge(self, evt):
        from petram.sol.evaluators import area_tri

        value = self.elps['Edge'] .GetValue()
        refine = int(value[6])
        data, data_x, battrs = self.eval_edge(mode='integ', refine=refine)
        if data is None:
            return

        ndim = data[0][0].shape[1]
        verts = np.hstack([v.flatten() for v, c, a in data]).flatten()
        cdata = np.hstack([c.flatten() for v, c, a in data]).flatten()
        verts = verts.reshape(-1, ndim)
        data = {'vertices': verts, 'data': cdata}

        if data_x is not None:
            cxdata = np.hstack([c.flatten() for v, c, a in data_x]).flatten()
            xverts = np.hstack([v.flatten() for v, c, a in data_x]).flatten()
            data['xvertices'] = xverts
            data['xdata'] = cxdata
        self.export_to_piScope_shell(data, 'edge_data')

    def get_attrs_field_Edge(self):
        return 2

    def eval_edge(self, mode='plot', refine=1):
        from petram.sol.evaluators import area_tri
        value = self.elps['Edge'] .GetValue()

        expr = str(value[0]).strip()
        expr_x = str(value[1]).strip()
        battrs = str(value[2])
        phys_path = value[3]
        if mode == 'plot':
            do_merge1 = value[5]
        else:
            do_merge1 = True
        average = value[7]

        exprs = [expr, expr_x] if expr_x != '' else [expr]
        data, void = self.evaluate_sol_edge(expr, battrs, phys_path,
                                            do_merge1, True,
                                            average=average,
                                            refine=refine,
                                            exprs=exprs)
        if data is None:
            return None, None, None

        if expr_x != '':
            data_x, void = self.evaluate_sol_edge(expr_x, battrs, phys_path,
                                                  do_merge1, True,
                                                  average=average,
                                                  refine=refine,
                                                  exprs=exprs)

            if data_x is None:
                return None, None, None
        else:
            data_x = None
        return data, data_x, battrs

    #
    #   Boundary value ('Bdr' tab)
    #
    '''
    ll = [['Expression', '', 0, {}],
                  ['Offset (x, y, z)', '0, 0, 0', 0, {}],
                  ['Boundary Index', text, 0, {}],
                  ['Physics', choices[0], 4, {'style':wx.CB_READONLY,
                                           'choices': choices}],
                  [None, False, 3, {"text":'dynamic extenstion'}],
                  [None, True, 3, {"text":'merge solutions'}],
                  ['Refine', 1, 104, s4],
                  [None, True, 3, {"text":'averaging'}],]
    '''
    @run_in_piScope_thread
    def onApplyBdr(self, evt):
        value = self.elps['Bdr'] .GetValue()
        expr = str(value[0]).strip()

        if value[7]:
            from ifigure.widgets.wave_viewer import WaveViewer
            cls = WaveViewer
        else:
            cls = None
        refine = int(value[9])
        use_pointfill = int(value[11]) > 1
        data, battrs = self.eval_bdr(mode='plot', refine=refine)
        if data is None:
            return

        scale = str(value[4]).strip()
        if scale.strip() != "":
            model = self.GetParent().model
            mfem_model = model.param.getvar('mfem_model')
            phys_path = value[6]
            phys_ns = mfem_model[str(phys_path)]._global_ns.copy()
            scale = eval(scale, {}, phys_ns)
        else:
            scale = 1.0

        self.post_threadend(self.make_plot_bdr, data, battrs, scale,
                            cls=cls, expr=expr,
                            use_pointfill=use_pointfill)

    def make_plot_bdr(
            self,
            data,
            battrs,
            scale,
            cls=None,
            expr='',
            use_pointfill=False):
        from ifigure.interactive import figure
        viewer = figure(viewer=cls)
        viewer.update(False)
        setup_figure(viewer, self.GetParent())
        viewer.suptitle(expr + ':' + str(battrs))

        dd = defaultdict(list)
        # regroup to sepparte triangles and quads.
        for k, datasets in enumerate(data):
            v, c, i = datasets  # verts, cdata, idata
            idx = i.shape[-1]
            dd[idx].append((k + 1, v, c, i))

        for key in dd.keys():
            kk, verts, cdata, idata = zip(*(dd[key]))
            offsets = np.hstack(
                (0, np.cumsum([len(c) for c in cdata], dtype=int)))[:-1]
            offsets_idx = np.hstack([np.zeros(len(a), dtype=int) + o
                                     for o, a in zip(offsets, idata)])
            array_idx = np.hstack([np.zeros(len(c), dtype=int) + k
                                   for k, c in zip(kk, cdata)])
            array_idx = array_idx + 1

            verts = np.vstack(verts)*scale
            cdata = np.hstack(cdata)
            idata = np.vstack(idata)
            idata = idata + np.atleast_2d(offsets_idx).transpose()

            if cls is None:
                obj = viewer.solid(verts, idata, array_idx=array_idx,
                                   cz=True, cdata=cdata.astype(float),
                                   shade='linear',
                                   use_pointfill=use_pointfill)
                obj.set_gl_hl_use_array_idx(True)
            else:
                obj = viewer.solid(verts, idata, array_idx=array_idx,
                                   cz=True, cdata=cdata, shade='linear',
                                   use_pointfill=use_pointfill)
                obj.set_gl_hl_use_array_idx(True)

        viewer.update(True)
        viewer.view('noclip')
        viewer.view('equal')
        viewer.update(False)
        ax = self.GetParent().get_axes()
        param = ax.get_axes3d_viewparam(ax._artists[0])
        ax2 = viewer.get_axes()
        ax2.set_axes3d_viewparam(param, ax2._artists[0])
        viewer.lighting(light=0.5)
        viewer.update(True)

    def onIntegBdr(self, evt):
        value = self.elps['Bdr'] .GetValue()
        expr = str(value[0]).strip()

        from petram.sol.evaluators import area_tri
        data, battrs = self.eval_bdr(mode='integ')
        if data is None:
            return

        integ = 0.0
        for verts, cdata, adata in data:
            v = verts[adata]
            c = cdata[adata, ...]
            area = area_tri(v)
            integ += np.sum(area * np.mean(c, 1))

        print("Area Ingegration")
        print("Expression : " + expr)
        print("Boundary Index :" + str(battrs))
        print("Value : " + str(integ))

    def onExportBdr(self, evt):
        from petram.sol.evaluators import area_tri
        data, battrs = self.eval_bdr(mode='integ')
        if data is None:
            return

        verts, cdata, adata = data[0]
        data = {'vertices': verts, 'data': cdata, 'index': adata}
        self.export_to_piScope_shell(data, 'bdr_data')

    @run_in_piScope_thread
    def onExportR1Bdr(self, evt):
        remote, base, subs = self.get_current_choices()

        cdata = []
        subdirs = []
        for s in subs:
            if s.strip() == '':
                continue
            if remote:
                self.config['cs_soldir'] = base
                self.config['cs_solsubdir'] = s
            else:
                self.local_soldir = base
                self.local_solsubdir = s
                self.load_sol_if_needed()

            data, battrs = self.eval_bdr(mode='integ')
            if data is None:
                assert False, "returned value is None ???"

            verts, cc, adata = data[0]
            cdata.append(cc)
            subdirs.append(s)

        data = {'vertices': verts, 'data': cdata, 'index': adata,
                'subdirs': subdirs}
        self.post_threadend(self.export_to_piScope_shell,
                            data, 'bdr_data')

    def onExportR2Bdr(self, evt):
        wx.CallAfter(
            dialog.showtraceback,
            parent=self,
            txt='Not Yet Implemented',
            title='Error',
            traceback='Exporing all time slice for frequency \ndomain analysis is not available')
        wx.CallAfter(self.set_title_no_status)

    def get_attrs_field_Bdr(self):
        return 1

    def eval_bdr(self, mode='plot', export_type=1, refine=1):
        from petram.sol.evaluators import area_tri
        value = self.elps['Bdr'] .GetValue()

        expr = str(value[0]).strip()
        battrs = str(value[5])
        phys_path = value[6]
        if mode == 'plot':
            do_merge1 = value[8]
            do_merge2 = True
        elif mode == 'integ':
            do_merge1 = True
            do_merge2 = False
        else:
            do_merge1 = False
            do_merge2 = False

        average = value[10]
        decimate = int(value[11])
        data, battrs2 = self.evaluate_sol_bdr(expr, battrs, phys_path,
                                              do_merge1, do_merge2,
                                              export_type=export_type,
                                              refine=refine,
                                              average=average,
                                              decimate=decimate)
        if data is None:
            return None, None

        uvw = (str(value[1]), str(value[2]), str(value[3]),)
        if len(uvw) == 3:
            for kk, expr in enumerate(uvw):
                if expr.strip() == '':
                    continue
                try:
                    u = float(expr.strip())
                    isfloat = True
                except BaseException:
                    isfloat = False
                    u, battrs2 = self.evaluate_sol_bdr(expr.strip(),
                                                       battrs, phys_path,
                                                       do_merge1, do_merge2,
                                                       export_type=export_type,
                                                       refine=refine,
                                                       average=average,
                                                       decimate=decimate)
                data = [list(x) for x in data]
                for k, datasets in enumerate(data):
                    if datasets[0].shape[1] == 2:
                        datasets[0] = np.hstack(
                            (datasets[0], np.zeros((datasets[0].shape[0], 1))))
                    elif datasets[0].shape[1] == 1:
                        datasets[0] = np.hstack(
                            (datasets[0], np.zeros((datasets[0].shape[0], 2))))

                    if isfloat:
                        datasets[0][:, kk] += u
                    else:
                        datasets[0][:, kk] += u[k][1]

        return data, battrs

    #
    #   PointCloud ('Points' tab)
    #
    @run_in_piScope_thread
    def onApplyPoints(self, evt):
        value = self.elps['Points'] .GetValue()

        expr = str(value[0]).strip()
        if value[4]:
            from ifigure.widgets.wave_viewer import WaveViewer
            cls = WaveViewer
        else:
            cls = None

        flag, dataset = self.call_eval_pointcloud(value)

        if flag > 0:
            ptx, data, attrs_out, attrs, pc_param = dataset
        else:
            return

        pc_mode = value[1][0]
        if pc_mode == 'XYZ':
            data = {'vertices': ptx, 'data': data, 'attrs': attrs_out}
            self.post_threadend(self.export_to_piScope_shell,
                                data,
                                'point_data')
        else:
            self.post_threadend(
                self.make_plot_point,
                data,
                attrs_out,
                attrs,
                pc_param,
                cls=cls,
                expr=expr)

    def onExportPoints(self, evt):
        value = self.elps['Points'] .GetValue()

        flag, dataset = self.call_eval_pointcloud(value)

        if flag > 0:
            ptx, data, attrs_out, attrs, pc_param = dataset
        else:
            return

        data = {'vertices': ptx, 'data': data, 'attrs': attrs_out}
        self.export_to_piScope_shell(data, 'point_data')

    @run_in_piScope_thread
    def onExportR1Points(self, evt):
        value = self.elps['Points'] .GetValue()

        remote, base, subs = self.get_current_choices()
        ret = []
        for s in subs:
            if s.strip() == '':
                continue
            if remote:
                self.config['cs_soldir'] = base
                self.config['cs_solsubdir'] = s
            else:
                self.local_soldir = base
                self.local_solsubdir = s
                self.load_sol_if_needed()

            flag, dataset = self.call_eval_pointcloud(value)
            if flag > 0:
                ptx, data, attrs_out, attrs, pc_param = dataset
                ret.append({'subdir': s, 'vertices': ptx,
                            'data': data, 'attrs': attrs_out})
            else:
                assert False, "pointcloud evaluation failed"

        self.post_threadend(self.export_to_piScope_shell,
                            ret, 'point_data')

    def onExportR2Points(self, evt):
        wx.CallAfter(
            dialog.showtraceback,
            parent=self,
            txt='Not Yet Implemented',
            title='Error',
            traceback='Exporing all time point for frequency \ndomain analysis is not available')
        wx.CallAfter(self.set_title_no_status)

    def call_eval_pointcloud(self, value):
        # def eval_pointcloud(self
        expr = str(value[0]).strip()
        attrs = str(value[2])
        phys_path = value[3]
        pc_mode = value[1][0]
        if pc_mode == 'XYZ':
            pc_value = value[1][1]
        elif pc_mode == 'Line':
            pc_value = value[1][2]
        else:
            return -1, None
        ptx, data, attrs_out, attrs, pc_param = self.eval_pointcloud(
            expr, attrs, phys_path, pc_mode, pc_value, mode='plot')
        if data is None:
            return -1, None
        return 1, (ptx, data, attrs_out, attrs, pc_param)

    def make_plot_point(
            self,
            data,
            attrs_out,
            attrs,
            param,
            cls=None,
            expr=False):
        value = self.elps['Points'] .GetValue()

        from ifigure.interactive import figure
        viewer = figure(viewer=cls)
        viewer.update(False)
        viewer.suptitle(expr + ':' + str(attrs))

        if param['pc_type'] == 'line':
            pc_param = param['pc_param']
            sp = np.array(pc_param[0])
            ep = np.array(pc_param[1])
            num = pc_param[2]
            ii = np.linspace(0, 1., num)
            ptx = np.vstack([sp * (1 - i) + ep * i for i in ii])
            if ptx.shape(-1) == 3:
                setup_figure(viewer, self.GetParent())
                viewer.plot(ptx[:, 0], ptx[:, 1], ptx[:, 2],
                            c=data.real, cz=True)
            else:
                viewer.plot(data.real)

        viewer.view('noclip')
        viewer.view('equal')
        ax = self.GetParent().get_axes()
        param = ax.get_axes3d_viewparam(ax._artists[0])
        ax2 = viewer.get_axes()
        ax2.set_axes3d_viewparam(param, ax2._artists[0])
        viewer.lighting(light=0.5)
        viewer.update(True)

    def process_abcd(self, abcd_txt, phys_ns):
        ll = {"YZ": _YZ((1, 0, 0., 0)),
              "XY": _XY((0., 0, 1., 0)),
              "ZX": _ZX((0., 1, 0., 0)),
              "yz": _YZ((1, 0, 0., 0)),
              "xy": _XY((0., 0, 1., 0)),
              "zx": _ZX((0., 1, 0., 0)), }
        # add all combinations
        ll["ZY"] = ll["YZ"]
        ll["zy"] = ll["YZ"]
        ll["YX"] = ll["XY"]
        ll["yx"] = ll["XY"]
        ll["XZ"] = ll["ZX"]
        ll["xz"] = ll["ZX"]

        abcd_value = eval(abcd_txt, ll, phys_ns)

        lens = [1, 1, 1, 1]
        for i in range(4):
            try:
                lens[i] = len(abcd_value[i])
            except TypeError:
                pass
        num_planes = max(lens)
        planes = [None]*4

        for i in range(4):
            if lens[i] == 1:
                planes[i] = [abcd_value[i]]*num_planes
            elif lens[i] == num_planes:
                planes[i] = abcd_value[i]
            else:
                if param is None:
                    wx.CallAfter(
                        dialog.showtraceback,
                        parent=self,
                        txt='Can not determin planes',
                        title='Error',
                        traceback='a, b, c, d has to have the same lenght when multiple planes are defined',)
                return None
        return planes

#    to time this routine, we turn on this decorator
#    from petram.debug import use_profiler
#    @use_profiler
    def eval_pointcloud(
            self,
            expr,
            attrs,
            phys_path,
            pc_mode,
            pc_value,
            mode='plot'):
        model = self.GetParent().model
        solfiles = self.get_model_soldfiles()

        mesh = model.variables.getvar('mesh')
        mfem_model = model.param.getvar('mfem_model')
        phys_ns = mfem_model[str(phys_path)]._global_ns.copy()

        ec = mesh.extended_connectivity
        v2s = ec['vol2surf']
        FaceOf, EdgeOf, PointOf = get_mapper(mesh)
        ll = {'FaceOf': FaceOf, 'EdgeOf': EdgeOf, 'PointOf': PointOf}

        if attrs == 'all':
            attrs = list(np.unique(mesh.GetAttributeArray()))

        elif attrs == 'visible':
            m = self.GetParent()
            battrs = []
            for name, child in m.get_axes(0).get_children():
                if name.startswith('face'):
                    battrs.extend(child.shown_component)
            battrs = list(set(battrs))
            if mesh.Dimenstion() == 3:
                attrs = [
                    k for k in v2s if set(
                        v2s[k]).intersection(battrs) == set(
                        v2s[k])]
            else:
                attrs = battrs

        elif attrs == 'hidden':
            m = self.GetParent()
            battrs = []
            sbattrs = []
            for name, child in m.get_axes(0).get_children():
                if name.startswith('face'):
                    battrs.extend(child.hidden_component)
                    sbattrs.extend(child.shown_component)
            battrs = list(set(battrs))
            if mesh.Dimenstion() == 3:
                attrs = [
                    k for k in v2s if set(
                        v2s[k]).intersection(sbattrs) != set(
                        v2s[k])]
            else:
                attrs = battrs

        else:
            try:
                attrs = list(np.atleast_1d(eval(attrs, ll, phys_ns)))
            except BaseException:
                traceback.print_exc()
                assert False, "invalid selection: " + attrs

        ll = {}
        if pc_mode == 'XYZ':
            xx = np.atleast_1d(eval(pc_value[0], ll, phys_ns))
            yy = np.atleast_1d(eval(pc_value[1], ll, phys_ns))
            zz = np.atleast_1d(eval(pc_value[2], ll, phys_ns))
            pc_param = {'pc_type': 'xyz',
                        'pc_param': np.stack([xx, yy, zz], -1)}
            ptx, data, attrs_out = self.evaluate_pointcloud(
                expr, attrs, phys_path, **pc_param)

            return ptx, data, attrs_out, attrs, pc_param

        elif pc_mode == 'Line':
            sp = tuple(np.atleast_1d(eval(pc_value[0], ll, phys_ns)))
            ep = tuple(np.atleast_1d(eval(pc_value[1], ll, phys_ns)))
            num = int(eval(pc_value[2], ll, phys_ns))
            pc_param = {'pc_type': 'line', 'pc_param': (sp, ep, num)}
            ptx, data, attrs_out = self.evaluate_pointcloud(
                expr, attrs, phys_path, **pc_param)

            return ptx, data, attrs_out, attrs, pc_param

        elif pc_mode == 'CutPlane':
            planes = self.process_abcd(pc_value[0], phys_ns)
            if planes is None:
                return None, None, None, None, None

            params = []
            e1 = list(np.atleast_1d(eval(pc_value[1], ll, phys_ns)))
            res = float(eval(pc_value[2], ll, phys_ns))
            from petram.helper.geom import find_cp_pc_parameter
            for abcd in zip(*planes):
                param = find_cp_pc_parameter(
                    mesh, abcd, e1, gsize=res, attrs=attrs)
                if param is not None:
                    param = (tuple(param["origin"]),
                             tuple(param["e1"]),
                             tuple(param["e2"]),
                             tuple(param["x"]),
                             tuple(param["y"]),)

                params.append(param)

            if len(params) == 0:
                wx.CallAfter(
                    dialog.showtraceback,
                    parent=self,
                    txt='No point is found',
                    title='Error',
                    traceback='Some cut plane does not intersect with geometry')
                return None, None, None, None, None

            ptx_list = []
            data_list = []
            attrs_out_list = []
            attrs_list = []
            pc_param_list = []

            for param in params:
                pc_param = {'pc_type': 'cutplane', 'pc_param': param}
                ptx, data, attrs_out = self.evaluate_pointcloud(
                    expr, attrs, phys_path, **pc_param)

                ptx_list.append(ptx)
                data_list.append(data)
                attrs_out_list.append(attrs_out)
                attrs_list.append(attrs)
                pc_param_list.append(pc_param)

            return ptx_list, data_list, attrs_out_list, attrs_list, pc_param_list

        else:
            return None, None, None, None, None

    def get_attrs_field_Points(self):
        return 2

    #
    #   Geometry Boundary ('GeomBdr' tab)
    #
    def onApplyGeomBdr(self, evt):
        x, y, z = self.eval_geombdr(mode='plot')

        value = self.elps['GeomBdr'] .GetValue()
        battrs = str(value[3])
        edge_only = bool(value[8])

        c1 = value[5][0]
        c2 = value[5][1]
        kwargs = {'facecolor': c1,
                  'edgecolor': c2, }
        if c2 == (0, 0, 0, 0):
            kwargs['linewidth'] = 0.

        from ifigure.interactive import figure
        v = figure()
        v.update(False)
        setup_figure(v, self.GetParent())
        v.suptitle('Boundary ' + str(battrs))
        for xdata, ydata, zdata in zip(x, y, z):
            verts = np.vstack((xdata[1], ydata[1], zdata[1])).transpose()
            adata = xdata[2]
            v.solid(verts, adata, **kwargs)

        v.update(True)
        v.view('noclip')
        v.view('equal')
        v.update(False)
        ax = self.GetParent().get_axes()
        param = ax.get_axes3d_viewparam(ax._artists[0])
        ax2 = v.get_axes()
        ax2.set_axes3d_viewparam(param, ax2._artists[0])
        v.lighting(light=0.5)
        v.update(True)

    def onExportGeomBdr(self, evt):
        from petram.sol.evaluators import area_tri
        x, y, z = self.eval_geombdr(mode='integ')
        # if data is None: return

        verts = np.dstack((x[0][1], y[0][1], z[0][1]))
        data = {'vertices': verts}
        self.export_to_piScope_shell(data, 'geom_data')

    def get_attrs_field_GeomBdr(self):
        return 3

    def eval_geombdr(self, mode='plot'):
        value = self.elps['GeomBdr'] .GetValue()
        cls = None
        expr_x = str(value[0]).strip()
        expr_y = str(value[1]).strip()
        expr_z = str(value[2]).strip()

        battrs = str(value[3])
        phys_path = value[4]
        edge_only = bool(value[8])
        if mode == 'plot':
            do_merge1 = value[6]
            do_merge2 = value[7]
        else:
            do_merge1 = True
            do_merge2 = False
        if edge_only:
            do_merge1 = False
            do_merge2 = False

        def call_eval_sol_bdr(expr, battrs=battrs, phys_path=phys_path,
                              do_merge1=do_merge1, do_merge2=do_merge2,
                              edge_only=edge_only):
            if str(expr).strip() != '':
                v, battrs = self.evaluate_sol_bdr(expr, battrs, phys_path,
                                                  do_merge1, do_merge2,
                                                  edge_only=edge_only)
            else:
                v = None
            return v
        x = call_eval_sol_bdr(expr_x)
        y = call_eval_sol_bdr(expr_y)
        z = call_eval_sol_bdr(expr_z)
        if x is None and y is None and z is None:
            return
        basedata = x
        if basedata is None:
            basedata = y
        if basedata is None:
            basedata = z

        zerodata = [(None, cdata * 0, adata) for verts, cdata, adata
                    in basedata]
        if x is None:
            x = zerodata
        if y is None:
            y = zerodata
        if z is None:
            z = zerodata
        return x, y, z

    #
    #   Arrow on Boundary ('Bdr(arrow)' tab)
    #
    @run_in_piScope_thread
    def onApplyBdrarrow(self, evt):
        u, v, w, battrs = self.eval_bdrarrow(mode='plot')

        value = self.elps['Bdr(arrow)'] .GetValue()

        expr_u = str(value[0]).strip()
        expr_v = str(value[1]).strip()
        expr_w = str(value[2]).strip()
        if value[5]:
            from ifigure.widgets.wave_viewer import WaveViewer
            cls = WaveViewer
        else:
            cls = None

        self.post_threadend(self.make_plot_bdrarrow, u, v, w, battrs, value,
                            expr_u=expr_u,
                            expr_v=expr_v,
                            expr_w=expr_w,
                            cls=cls)

    def make_plot_bdrarrow(self, u, v, w, battrs, value,
                           expr_u='', expr_v='', expr_w='',
                           cls=None):

        from ifigure.interactive import figure
        viewer = figure(viewer=cls)
        viewer.update(False)
        setup_figure(viewer, self.GetParent())
        viewer.suptitle(
            '[' + ','.join((expr_u, expr_v, expr_w)) + '] : ' + str(battrs))

        allxyz = np.vstack([udata[0] for udata in u])
        dx = np.max(allxyz[:, 0]) - np.min(allxyz[:, 0])
        if allxyz.shape[1] > 1:
            dy = np.max(allxyz[:, 1]) - np.min(allxyz[:, 1])
        else:
            dy = dx * 0.
        if allxyz.shape[1] > 2:
            dz = np.max(allxyz[:, 2]) - np.min(allxyz[:, 2])
        else:
            dz = dy * 0.
        length = np.max((dx, dy, dz)) / 20.

        for udata, vdata, wdata in zip(u, v, w):
            xyz = udata[0]

            u = udata[1]
            v = vdata[1]
            w = wdata[1]

            ll = np.min([xyz.shape[0] - 1, int(value[7])])
            idx = np.linspace(0, xyz.shape[0] - 1, ll).astype(int)

            x = xyz[idx, 0]
            if xyz.shape[1] > 1:
                y = xyz[idx, 1]
            else:
                y = x * 0.
            if xyz.shape[1] > 2:
                z = xyz[idx, 2]
            else:
                z = x * 0.

            viewer.quiver3d(x, y, z, u[idx], v[idx], w[idx],
                            length=length)

        viewer.update(True)
        viewer.view('noclip')
        viewer.view('equal')
        viewer.update(False)
        ax = self.GetParent().get_axes()
        param = ax.get_axes3d_viewparam(ax._artists[0])
        ax2 = viewer.get_axes()
        ax2.set_axes3d_viewparam(param, ax2._artists[0])
        viewer.lighting(light=0.5)
        viewer.update(True)

    def onExportBdrarrow(self, evt):
        u, v, w, battrs = self.eval_bdrarrow(mode='export')
        udata = u[0][1]
        vdata = v[0][1]
        wdata = w[0][1]
        verts = v[0][0]
        xyz = np.mean(verts, 1)
        u = np.mean(udata, 1)
        v = np.mean(vdata, 1)
        w = np.mean(wdata, 1)
        data = {'x': xyz[:, 0],
                'y': xyz[:, 1],
                'z': xyz[:, 2],
                'u': u,
                'v': v,
                'w': w}
        self.export_to_piScope_shell(data, 'arrow_data')

    def get_attrs_field_Bdrarrow(self):
        return 3

    def eval_bdrarrow(self, mode='plot'):
        value = self.elps['Bdr(arrow)'] .GetValue()
        cls = None
        expr_u = str(value[0]).strip()
        expr_v = str(value[1]).strip()
        expr_w = str(value[2]).strip()

        battrs = str(value[3])
        phys_path = value[4]
        if mode == 'plot':
            do_merge1 = value[6]
            do_merge2 = False
        else:
            do_merge1 = True
            do_merge2 = False

        def call_eval_sol_bdr(expr, battrs=battrs, phys_path=phys_path,
                              do_merge1=do_merge1, do_merge2=do_merge2):
            if str(expr).strip() != '':
                v, battrs = self.evaluate_sol_bdr(expr, battrs, phys_path,
                                                  do_merge1, do_merge2)
            else:
                v = None
                battrs = None
            return v, battrs

        u, ubattrs = call_eval_sol_bdr(expr_u)
        v, vbattrs = call_eval_sol_bdr(expr_v)
        w, wbattrs = call_eval_sol_bdr(expr_w)
        if u is None and v is None and w is None:
            return

        basedata = u
        battrs = ubattrs
        if basedata is None:
            basedata = v
            battrs = vbattrs
        if basedata is None:
            basedata = w
            battrs = wbattrs

        zerodata = [(verts, cdata * 0, adata)
                    for verts, cdata, adata in basedata]
        if u is None:
            u = zerodata
        if v is None:
            v = zerodata
        if w is None:
            w = zerodata
        return u, v, w, battrs
    #
    #   Slice plane ('Slice' tab)
    #

    @run_in_piScope_thread
    def onApplySlice(self, evt):
        self.onSliceCommon(evt)

    def onExportSlice(self, evt):
        self.onSliceCommon(evt, mode='export')

    @run_in_piScope_thread
    def onxportR1Slice(self, evt):
        remote, base, subs = self.get_current_choices()

        dataset = []
        for s in subs:
            if s.strip() == '':
                continue
            if remote:
                self.config['cs_soldir'] = base
                self.config['cs_solsubdir'] = s
            else:
                self.local_soldir = base
                self.local_solsubdir = s
                self.load_sol_if_needed()

            data = self.onSliceCommon(evt, mode='export_return')
            if data is None:
                assert False, "returned value is None ???"

            dataset.append({"subdir", s, "data", data})

        self.post_threadend(self.export_to_piScope_shell,
                            dataset, 'slice_data')

    def onExportR2Slice(self, evt):
        wx.CallAfter(
            dialog.showtraceback,
            parent=self,
            txt='Not Yet Implemented',
            title='Error',
            traceback='Exporing all time slice for frequency \ndomain analysis is not available')
        wx.CallAfter(self.set_title_no_status)

    def onSliceCommon(self, evt, mode='plot'):

        value = self.elps['Slice'] .GetValue()

        expr = str(value[0]).strip()

        if value[4]:
            from ifigure.widgets.wave_viewer import WaveViewer
            cls = WaveViewer
        else:
            cls = None

        if value[1][0] == 'Triangles':
            phys_path = value[3]
            data, battrs = self.eval_slice(phys_path, mode='plot')
            if data is None:
                wx.CallAfter(self.set_title_no_status)
                return
            if mode == 'export':
                data = {'data': data}
                self.export_to_piScope_shell(data, 'slice_data')
            elif mode == 'export_return':
                data = {'data': data}
                return data
            else:
                self.post_threadend(self.make_plot_slice,
                                    data, battrs, cls=cls, expr=expr)
                return

        elif value[1][0] == 'Grid':
            attrs = str(value[2])
            phys_path = value[3]
            pc_mode = 'CutPlane'
            pc_value = value[1][2]
            ptx_list, data_list, attrs_out_list, attrs_list, pc_param_list = self.eval_pointcloud(
                expr, attrs, phys_path, pc_mode, pc_value, mode='plot')

            num = 1
            dataset = []

            for ptx, data, attrs_out, attrs, pc_param in zip(ptx_list, data_list, attrs_out_list,
                                                             attrs_list, pc_param_list):

                if data is None:
                    continue

                pc_param = pc_param['pc_param']
                im_center = pc_param[0]
                im_axes = (pc_param[1], pc_param[2])
                midx = (pc_param[3][0] + pc_param[3][1]) / 2.0
                midy = (pc_param[4][0] + pc_param[4][1]) / 2.0
                xmin, xmax, xsize = pc_param[3]
                ymin, ymax, ysize = pc_param[4]

                x = np.linspace(xmin, xmax, int((xmax - xmin) / xsize))
                y = np.linspace(ymin, ymax, int((ymax - ymin) / ysize))

                suffix = "_plane"+str(num)
                if mode == 'export':
                    data = {'data'+suffix: data, 'x'+suffix: x, 'y'+suffix: y,
                            'im_axes'+suffix: im_axes, 'im_center'+suffix: im_center}
                    self.export_to_piScope_shell(data, 'slice_data')
                elif mode == 'export_return':
                    data = {'data'+suffix: data, 'x'+suffix: x, 'y'+suffix: y,
                            'im_axes'+suffix: im_axes, 'im_center'+suffix: im_center}
                    dataset.append(data)
                else:
                    data = {'data': data, 'x': x, 'y': y,
                            'im_axes': im_axes, 'im_center': im_center,
                            'attrs_out': attrs_out,
                            'attrs': attrs}
                    dataset.append(data)
                num = num+1

            if mode == 'export_return':
                return dataset
            elif mode == 'plot':
                self.post_threadend(
                    self.make_plot_pc_slice,
                    dataset,
                    cls=cls,
                    expr=expr)

    def make_plot_slice(self, dataset, battrsset, cls=None, expr=''):
        from ifigure.interactive import figure
        v = figure(viewer=cls)
        v.update(False)
        setup_figure(v, self.GetParent())

        first = True
        for data,  battrs in zip(dataset, battrsset):
            if first:
                v.suptitle(expr + ':' + str(battrs))
                first = False
            for verts, cdata, adata in data:
                if cls is None:
                    v.solid(verts, adata, cz=True, cdata=cdata.astype(float),
                            shade='linear')
                else:
                    v.solid(verts, adata, cz=True, cdata=cdata, shade='linear')

        v.update(True)
        v.view('noclip')
        v.view('equal')
        v.update(False)
        ax = self.GetParent().get_axes()
        param = ax.get_axes3d_viewparam(ax._artists[0])
        ax2 = v.get_axes()
        ax2.set_axes3d_viewparam(param, ax2._artists[0])
        v.lighting(light=0.5)
        v.update(True)

    def make_plot_pc_slice(self, dataset, cls=None, expr=False):

        from ifigure.interactive import figure
        viewer = figure(viewer=cls)
        viewer.update(False)
        setup_figure(viewer, self.GetParent())
        viewer.threed('on')

        first = True
        for d in dataset:
            data = d['data']
            attrs_out = d['attrs_out']
            attrs = d['attrs']
            x = d['x']
            y = d['y']
            im_axes = d['im_axes']
            im_center = d['im_center']

            if first:
                viewer.suptitle(expr + ':' + str(attrs))
                first = False
            if cls is None:
                data = np.ma.masked_array(
                    data.real, mask=np.in1d(
                        attrs_out, attrs, invert=True))
            else:
                data = np.ma.masked_array(
                    data, mask=np.in1d(
                        attrs_out, attrs, invert=True))
            viewer.image(x, y, data, im_axes=im_axes, im_center=im_center)

            viewer.view('noclip')
            viewer.view('equal')
            ax = self.GetParent().get_axes()
            param = ax.get_axes3d_viewparam(ax._artists[0])
            ax2 = viewer.get_axes()
            ax2.set_axes3d_viewparam(param, ax2._artists[0])
            viewer.lighting(light=0.5)

        viewer.update(True)

    def get_attrs_field_Slice(self):
        return 2

    def eval_slice(self, phys_path, mode='plot'):
        model = self.GetParent().model
        mfem_model = model.param.getvar('mfem_model')
        phys_ns = mfem_model[str(phys_path)]._global_ns.copy()

        value = self.elps['Slice'] .GetValue()

        expr = str(value[0]).strip()

        planes = self.process_abcd(str(value[1][1][0]), phys_ns)

        attrs = str(value[2])
        phys_path = value[3]
        if mode == 'plot':
            do_merge1 = value[5]
            do_merge2 = False
        else:
            do_merge1 = True
            do_merge2 = False

        dataset = []
        attrsset = []
        for plane in zip(*planes):
            data, battrs = self.evaluate_sol_slice(expr, attrs, plane, phys_path,
                                                   do_merge1, do_merge2)
            if data is not None:
                dataset.append(data)
                attrsset.append(battrs)

        if len(dataset) == 0:
            return None, None
        return dataset, attrsset

    '''
    integral
    '''
    @run_in_piScope_thread
    def onApplyIntegral(self, evt):
        expr, value, kind, idx, order = self.eval_integral()
        if value is not None:
            data = {'value': value,
                    'kind': kind,
                    'order': order,
                    'idx': idx,
                    'expr': expr, }
            self.post_threadend(self.export_to_piScope_shell,
                                data, 'integral_data')
            self.post_threadend(print,
                                "Integrated value", data['value'])

    def get_attrs_field_Integral(self):
        return 2

    def eval_integral(self):
        value = self.elps['Integral'] .GetValue()
        expr = str(value[0]).strip()
        kind = str(value[1]).strip()
        attrs = str(value[2])
        order = int(value[3])
        phys_path = str(value[4]).strip()

        value = self.evaluate_sol_integral(expr, kind, attrs, order, phys_path)
        return expr, value, kind, attrs, order
    '''
    probe
    '''
    @run_in_piScope_thread
    def onApplyProbe(self, evt):
        value = self.elps['Probe'] .GetValue()
        expr = str(value[0]).strip()
        xexpr = str(value[1]).strip()

        xdata, data = self.eval_probe(mode='plot')
        if data is None:
            wx.CallAfter(self.set_title_no_status)
            return
        self.post_threadend(
            self.make_plot_probe, (xdata, data), expr=expr, xexpr=xexpr)

    def onExportProbe(self, evt):
        value = self.elps['Probe'] .GetValue()
        xdata, data = self.eval_probe(mode='plot')

        data = {'xdata': xdata, 'data': data}
        self.export_to_piScope_shell(data, 'probe_data')

    def make_plot_probe(self, data, expr='', xexpr='', cls=None):
        from ifigure.interactive import figure
        v = figure(viewer=cls)
        v.update(False)
        v.suptitle(expr)
        if len(xexpr) != 0:
            v.xlabel(xexpr)
            v.plot(data[0], data[1])
        else:
            v.plot(data[1])
        v.update(True)

    def eval_probe(self, mode='plot'):
        value = self.elps['Probe'] .GetValue()
        expr = str(value[0]).strip()
        xexpr = str(value[1]).strip()
        phys_path = value[2]
        xdata, data = self.evaluate_sol_probe(expr, xexpr, phys_path)
        return xdata, data

    #
    #   common routines
    #
    def evaluate_sol_edge(self, expr, battrs, phys_path, do_merge1, do_merge2,
                          **kwargs):
        '''
        evaluate sol using boundary evaluator
        '''
        model = self.GetParent().model
        solfiles = self.get_model_soldfiles()
        mfem_model = model.param.getvar('mfem_model')
        phys_ns = mfem_model[str(phys_path)]._global_ns.copy()

        if solfiles is None:
            wx.CallAfter(dialog.showtraceback, parent=self,
                         txt='Solution does not exist',
                         title='Error',
                         traceback='')
            wx.CallAfter(self.set_title_no_status)
            return None, None
        mesh = model.variables.getvar('mesh')
        if mesh is None:
            return

        FaceOf, EdgeOf, PointOf = get_mapper(mesh)
        ll = {'FaceOf': FaceOf, 'EdgeOf': EdgeOf, 'PointOf': PointOf}

        battrs = str(battrs).strip()
        if battrs.lower() == 'all':
            battrs = list(mesh.extended_connectivity['line2vert'])
        else:
            try:
                battrs = list(np.atleast_1d(eval(battrs, ll, phys_ns)))
            except BaseException:
                traceback.print_exc()
                assert False, "invalid selection: " + battrs

        from petram.sol.evaluators import build_evaluator

        average = kwargs.pop('average', True)
        if average:
            key, name = 'Edge', 'EdgeNodal'
        else:
            key, name = 'NCEdge', 'NCEdge'

        if key in self.evaluators:
            try:
                self.evaluators[key].validate_evaluator(name,
                                                        battrs,
                                                        solfiles)
            except IOError:
                dprint1("IOError detected setting failed=True")
                self.evaluators[key].failed = True

        if (key not in self.evaluators or
                self.evaluators[key].failed):
            if key in self.evaluators:
                self.evaluators[key].terminate_all()
            try:
                self.evaluators[key] = build_evaluator(battrs,
                                                       mfem_model,
                                                       solfiles,
                                                       name=name,
                                                       config=self.config)
            except:
                wx.CallAfter(dialog.showtraceback, parent=self,
                             txt='Failed to build evaluator',
                             title='Error',
                             traceback=''.join(traceback.format_exception_only(
                                 sys.exc_info()[0], sys.exc_info()[1])))
                wx.CallAfter(self.set_title_no_status)
                return None, None

            self.evaluators[key].validate_evaluator(name,
                                                    battrs,
                                                    solfiles, isFirst=True)

        try:
            self.evaluators[key].set_phys_path(phys_path)
            return self.evaluators[key].eval(expr, do_merge1, do_merge2,
                                             **kwargs)
        except BaseException:
            traceback.print_exc()
            wx.CallAfter(dialog.showtraceback, parent=self,
                         txt='Failed to evauate expression',
                         title='Error',
                         traceback=''.join(traceback.format_exception_only(
                             sys.exc_info()[0], sys.exc_info()[1])))
            wx.CallAfter(self.set_title_no_status)
        return None, None

    def evaluate_sol_bdr(self, expr, battrs, phys_path, do_merge1, do_merge2,
                         **kwargs):
        '''
        evaluate sol using boundary evaluator
        '''
        model = self.GetParent().model
        solfiles = self.get_model_soldfiles()
        mfem_model = model.param.getvar('mfem_model')
        phys_ns = mfem_model[str(phys_path)]._global_ns.copy()

        if solfiles is None:
            wx.CallAfter(dialog.showtraceback, parent=self,
                         txt='Solution does not exist',
                         title='Error',
                         traceback='')
            wx.CallAfter(self.set_title_no_status)
            return None, None
        mesh = model.variables.getvar('mesh')
        if mesh is None:
            return
        FaceOf, EdgeOf, PointOf = get_mapper(mesh)
        ll = {'FaceOf': FaceOf, 'EdgeOf': EdgeOf, 'PointOf': PointOf}

        if battrs == 'all':
            battrs = list(mesh.extended_connectivity['surf2line'])
        elif battrs == 'visible':
            m = self.GetParent()
            battrs = []
            for name, child in m.get_axes(0).get_children():
                if name.startswith('face'):
                    battrs.extend(child.shown_component)
            battrs = list(set(battrs))
        elif battrs == 'hidden':
            m = self.GetParent()
            battrs = []
            for name, child in m.get_axes(0).get_children():
                if name.startswith('face'):
                    battrs.extend(child.hidden_component)
            battrs = list(set(battrs))
        else:
            try:
                battrs = list(np.atleast_1d(eval(battrs, ll, phys_ns)))
            except BaseException:
                traceback.print_exc()
                assert False, "invalid selection: " + battrs

            #battrs = [x+1 for x in range(mesh.bdr_attributes.Size())]

        average = kwargs.pop('average', True)
        decimate = kwargs.get('decimate', 1)

        from petram.sol.evaluators import build_evaluator
        if average:
            key, name = 'Bdr', 'BdrNodal'
        else:
            key, name = 'NCFace', 'NCFace'

        if key in self.evaluators:
            try:
                self.evaluators[key].validate_evaluator(name, battrs, solfiles,
                                                        decimate=decimate)
            except IOError:
                dprint1("IOError detected setting failed=True")
                self.evaluators[key].failed = True

        if (key not in self.evaluators or
                self.evaluators[key].failed):
            if key in self.evaluators:
                self.evaluators[key].terminate_all()
            try:
                self.evaluators[key] = build_evaluator(battrs,
                                                       mfem_model,
                                                       solfiles,
                                                       name=name,
                                                       config=self.config,
                                                       decimate=decimate)
            except:
                wx.CallAfter(dialog.showtraceback, parent=self,
                             txt='Failed to build evaluator',
                             title='Error',
                             traceback=''.join(traceback.format_exception_only(
                                 sys.exc_info()[0], sys.exc_info()[1])))
                wx.CallAfter(self.set_title_no_status)
                return None, None

            self.evaluators[key].validate_evaluator(name,
                                                    battrs,
                                                    solfiles, isFirst=True,
                                                    decimate=decimate)

        try:
            self.evaluators[key].set_phys_path(phys_path)
            return self.evaluators[key].eval(expr, do_merge1, do_merge2,
                                             **kwargs)
        except BaseException:
            wx.CallAfter(dialog.showtraceback, parent=self,
                         txt='Failed to evauate expression',
                         title='Error',
                         traceback=traceback.format_exc())
            wx.CallAfter(self.set_title_no_status)
        return None, None

    def evaluate_pointcloud(self, expr, attrs, phys_path, **kwargs):
        '''
        evaluate sol using boundary evaluator
        '''
        model = self.GetParent().model
        solfiles = self.get_model_soldfiles()
        mfem_model = model.param.getvar('mfem_model')
        phys_ns = mfem_model[str(phys_path)]._global_ns.copy()

        if solfiles is None:
            wx.CallAfter(dialog.showtraceback, parent=self,
                         txt='Solution does not exist',
                         title='Error',
                         traceback='')
            wx.CallAfter(self.set_title_no_status)
            return None, None, None
        mesh = model.variables.getvar('mesh')
        if mesh is None:
            return None, None, None

        from petram.sol.evaluators import build_evaluator
        key, name = 'Points', 'Pointcloud'

        if key in self.evaluators:
            try:
                self.evaluators[key].validate_evaluator(
                    name, attrs, solfiles, **kwargs)

            except IOError:
                dprint1("IOError detected setting failed=True")
                self.evaluators[key].failed = True

        if (key not in self.evaluators or
                self.evaluators[key].failed):
            if key in self.evaluators:
                self.evaluators[key].terminate_all()

            try:
                self.evaluators[key] = build_evaluator(attrs,
                                                       mfem_model,
                                                       solfiles,
                                                       name=name,
                                                       config=self.config,
                                                       **kwargs)
            except:
                wx.CallAfter(dialog.showtraceback, parent=self,
                             txt='Failed to build evaluator',
                             title='Error',
                             traceback=''.join(traceback.format_exception_only(
                                 sys.exc_info()[0], sys.exc_info()[1])))
                wx.CallAfter(self.set_title_no_status)
                return None, None

            self.evaluators[key].validate_evaluator(name,
                                                    attrs,
                                                    solfiles, isFirst=True,
                                                    **kwargs)

        try:
            self.evaluators[key].set_phys_path(phys_path)
            return self.evaluators[key].eval_pointcloud(expr)
        except BaseException:
            wx.CallAfter(dialog.showtraceback, parent=self,
                         txt='Failed to evauate expression',
                         title='Error',
                         traceback=traceback.format_exc())
            wx.CallAfter(self.set_title_no_status)
        return None, None, None

    def evaluate_sol_slice(self, expr, attrs, plane, phys_path, do_merge1,
                           do_merge2):
        '''
        evaluate sol using slice evaluator
        '''
        model = self.GetParent().model
        solfiles = self.get_model_soldfiles()
        mfem_model = model.param.getvar('mfem_model')
        phys_ns = mfem_model[str(phys_path)]._global_ns.copy()
        ll = {}

        if solfiles is None:
            wx.CallAfter(dialog.showtraceback, parent=self,
                         txt='Solution does not exist',
                         title='Error',
                         traceback='')
            wx.CallAfter(self.set_title_no_status)
            return None, None
        mesh = model.variables.getvar('mesh')
        if mesh is None:
            return None, None

        if mesh.SpaceDimension() != 3:
            wx.CallAfter(dialog.showtraceback, parent=self,
                         txt='Triangle Slice plot works only for 3D Tet mesh',
                         title='Error',
                         traceback='')
            wx.CallAfter(self.set_title_no_status)
            return None, None
        if attrs != 'all':
            try:
                attrs = list(np.atleast_1d(eval(attrs, ll, phys_ns)))
            except BaseException:
                traceback.print_exc()
                assert False, "Failed to evaluate attrs " + attrs
        else:
            attrs = list(mesh.extended_connectivity['vol2surf'])
            #attrs = [x+1 for x in range(mesh.attributes.Size())]

        if 'Slice' in self.evaluators:
            try:
                self.evaluators['Slice'].validate_evaluator(
                    'Slice', attrs, solfiles, plane=plane)
            except IOError:
                dprint1("IOError detected setting failed=True")
                self.evaluators['Slice'].failed = True

        from petram.sol.evaluators import build_evaluator
        if ('Slice' not in self.evaluators or
                self.evaluators['Slice'].failed):

            if 'Slice' in self.evaluators:
                self.evaluators['Slice'].terminate_all()

            try:
                self.evaluators['Slice'] = build_evaluator(attrs,
                                                           mfem_model,
                                                           solfiles,
                                                           name='Slice',
                                                           config=self.config,
                                                           plane=plane)
            except:
                wx.CallAfter(dialog.showtraceback, parent=self,
                             txt='Failed to build evaluator',
                             title='Error',
                             traceback=''.join(traceback.format_exception_only(
                                 sys.exc_info()[0], sys.exc_info()[1])))
                wx.CallAfter(self.set_title_no_status)
                return None, None

            self.evaluators['Slice'].validate_evaluator('Slice', attrs,
                                                        solfiles, isFirst=True,
                                                        plane=plane)

        try:
            self.evaluators['Slice'].set_phys_path(phys_path)
            # we do this in order to catch a communication error, which return
            # None
            ret1, ret2 = self.evaluators['Slice'].eval(
                expr, do_merge1, do_merge2)
            return ret1, ret2
        except BaseException:
            wx.CallAfter(dialog.showtraceback,
                         parent=self,
                         txt='Failed to evauate expression',
                         title='Error',
                         traceback=traceback.format_exc())

            wx.CallAfter(self.set_title_no_status)
        return None, None

    def evaluate_sol_integral(self, expr, kind, attrs, order, phys_path):
        model = self.GetParent().model
        solfiles = self.get_model_soldfiles()
        mfem_model = model.param.getvar('mfem_model')

        phys_ns = mfem_model[str(phys_path)]._global_ns.copy()
        mesh = model.variables.getvar('mesh')

        if attrs != 'all':
            try:
                attrs = list(np.atleast_1d(eval(attrs, {}, phys_ns)))
            except BaseException:
                traceback.print_exc()
                assert False, "Failed to evaluate attrs " + attrs
        else:
            if mesh.Dimension() == 3:
                if kind == 'Domain':
                    attrs = list(mesh.extended_connectivity['vol2surf'])
                else:
                    attrs = list(mesh.extended_connectivity['surf2line'])
            elif mesh.Dimension() == 2:
                if kind == 'Domain':
                    attrs = list(mesh.extended_connectivity['surf2line'])
                else:
                    attrs = list(mesh.extended_connectivity['line2vert'])
            elif mesh.Dimension() == 1:
                attrs = list(mesh.extended_connectivity['line2vert'])
            else:
                assert False, "unsupported mesh dimension"

        from petram.sol.evaluators import build_evaluator

        if 'Integral' in self.evaluators:
            self.evaluators['Integral'].terminate_all()

        try:
            self.evaluators['Integral'] = build_evaluator(attrs,
                                                          mfem_model,
                                                          solfiles,
                                                          name='Integral',
                                                          config=self.config)
        except:
            wx.CallAfter(dialog.showtraceback, parent=self,
                         txt='Failed to build evaluator',
                         title='Error',
                         traceback=''.join(traceback.format_exception_only(
                             sys.exc_info()[0], sys.exc_info()[1])))
            wx.CallAfter(self.set_title_no_status)
            return None

        try:
            self.evaluators['Integral'].set_phys_path(phys_path)
            return self.evaluators['Integral'].eval_integral(expr,
                                                             kind=kind,
                                                             attrs=attrs,
                                                             order=order)
        except BaseException:
            wx.CallAfter(dialog.showtraceback,
                         parent=self,
                         txt='Failed to evauate expression (integral)',
                         title='Error',
                         traceback=traceback.format_exc())

            wx.CallAfter(self.set_title_no_status)
        return None

    def evaluate_sol_probe(self, expr, xexpr, phys_path):
        model = self.GetParent().model
        solfiles = None   # probe does not load solfile (GridFunction)
        mfem_model = model.param.getvar('mfem_model')

        attrs = [1]

        from petram.sol.evaluators import build_evaluator

        if 'Probe' in self.evaluators:
            self.evaluators['Probe'].terminate_all()

        try:
            self.evaluators['Probe'] = build_evaluator(attrs,
                                                       mfem_model,
                                                       solfiles,
                                                       name='Probe',
                                                       config=self.config)
        except:
            wx.CallAfter(dialog.showtraceback, parent=self,
                         txt='Failed to build evaluator',
                         title='Error',
                         traceback=''.join(traceback.format_exception_only(
                             sys.exc_info()[0], sys.exc_info()[1])))
            wx.CallAfter(self.set_title_no_status)
            return None, None

        try:
            if not self.config['use_cs']:
                probes = self.local_sols[0:2]
            else:
                probes = self.remote_sols[0:2]

            self.evaluators['Probe'].set_phys_path(phys_path)
            data = self.evaluators['Probe'].eval_probe(expr, xexpr, probes)
            return data[1], np.transpose(data[2])
        except BaseException:
            wx.CallAfter(dialog.showtraceback,
                         parent=self,
                         txt='Failed to evauate expression (probe)',
                         title='Error',
                         traceback=traceback.format_exc())

            wx.CallAfter(self.set_title_no_status)
        return None, None

    #
    #   utilites
    #
    def export_to_piScope_shell(self, data, dataname):
        import wx
        import ifigure.widgets.dialog as dialog

        app = wx.GetApp().TopWindow
        app.shell.lvar[dataname] = data
        app.shell.SendShellEnterEvent()
        ret = dialog.message(app, dataname + ' is exported', 'Export', 0)

    def get_model_soldfiles(self):
        model = self.GetParent().model
        solfiles = model.variables.getvar('solfiles')
        soldir = model.variables.getvar('remote_soldir')

        if not self.config['use_cs']:
            return solfiles
        else:
            soldir = os.path.join(soldir, self.config["cs_solsubdir"])
            return soldir



from collections import defaultdict
import numpy as np

def find_vertex(mesh, bb_edges):
    

    corners = {}
    for key in bb_edges:
        iv = np.hstack([mesh.GetEdgeVertices(ie) for ie in bb_edges[key]])
        seen = defaultdict(int)
        for iiv in iv:
            seen[iiv] += 1
        corners[key] = [kk for kk in seen if seen[kk]==1]

    ivert = np.unique(np.hstack([corners[key]
                                 for key in corners])).astype(int, copy=False)
    return corners, ivert
        



'''
   find_loop_par

   (parallel) find edge index surrouding a face

'''
import numpy as np
from collections import defaultdict

def find_loop_ser(mesh, *face):
    '''
    find_loop_ser(mesh, 1)                 # loop around boundary index = 1
    find_loop_ser(mesh, [1, 2, 3])           # loop around boundary made by union of index = 1,2,3
    find_loop_ser(mesh, [1,2,3], [4, 5]) # loop made by two set of boundaries 
                                              # (1,2,3) and (4, 5)

    '''
    if len(face) == 1:
        face = face[0]
        faces = np.atleast_1d(face)
        battrs = mesh.GetBdrAttributeArray()
        bidx = np.where(np.in1d(battrs, faces))[0]
        edges = [mesh.GetBdrElementEdges(i) for i in bidx]
        iedges = sum([e1[0] for e1 in edges], [])
        dirs =  sum([e1[1] for e1 in edges], [])
    else:
        battrs = mesh.GetBdrAttributeArray()
        
        bidx1 = np.where(np.in1d(battrs, face[0]))[0]
        edges1 = np.unique(np.hstack([mesh.GetBdrElementEdges(i)[0] for i in bidx1]))
        bidx2 = np.where(np.in1d(battrs, face[1]))[0]
        edges2 = np.unique(np.hstack([mesh.GetBdrElementEdges(i)[0] for i in bidx2]))

        iedges = np.intersect1d(edges1, edges2)

        #dirs = {}
        #for i in bidx2:
        #     print(mesh.GetBdrElementEdges(i))
        #     for ei, dir in zip(*mesh.GetBdrElementEdges(i)):
        #         if ei in dirs and  dirs[ei] != dir:
        #             print("orientation is not obtained by this way")
        #         dirs[ei] = dir
        #dirs = [dirs[i] for i in iedges]

    seens = defaultdict(int)
    seendirs  = defaultdict(int)
    
    for k, ie in enumerate(iedges):
        seens[ie] = seens[ie] + 1
        seendirs[ie] = dirs[k]
    seens.default_factory = None
     
    idx = []
    signs = []
    for k in seens.keys():
        if seens[k] == 1:
            idx.append(k)
            signs.append(seendirs[k])

    ll = {tuple(mesh.GetEdgeVertices(i)):i for i in idx}

    from petram.helper.geom import connect_pairs

    loop = connect_pairs(list(ll))

    dirs = {}
    for i in range(len(loop)-1):
        if (loop[i], loop[i+1]) in ll: dirs[(loop[i], loop[i+1])] = 1
        if (loop[i+1], loop[i]) in ll: dirs[(loop[i+1], loop[i])] = -1

    signs = [dirs[key] for key in ll.keys()]
    idx = [ll[key] for key in ll.keys()]    
        
    print("here",idx, signs)
    return idx, signs
    
def find_loop_par(pmesh, *face):
    '''
    find_loop_ser(mesh, 1)          # loop around boundary index = 1
    find_loop_ser(mesh, [1, 2, 3])    # loop around boundary made by union of index = 1,2,3
    '''
    
    import mfem.par as mfem
    from mpi4py import MPI

    myid = MPI.COMM_WORLD.rank
    nprc = MPI.COMM_WORLD.size   
    comm  = MPI.COMM_WORLD

    if nprc == 1:
        return find_loop_ser(pmesh, face)

    from mfem.common.mpi_debug import nicePrint
    from petram.helper.mpi_recipes import allgather, allgather_vector, gather_vector    

    battrs = pmesh.GetBdrAttributeArray()

    face = face[0]
    faces = np.atleast_1d(face)
    bidx = np.where(np.in1d(battrs,faces))[0]
    
    offset_e = np.hstack([0, np.cumsum(allgather(pmesh.GetNEdges()))])
    
    edges = [pmesh.GetBdrElementEdges(i) for i in bidx]
    iedges = np.array(sum([e1[0] for e1 in edges], []), dtype=int) + offset_e[myid]
    dirs =  np.array(sum([e1[1] for e1 in edges], []), dtype=int)

    from petram.mesh.mesh_utils import distribute_shared_entity

    shared_info = distribute_shared_entity(pmesh)

    keys = shared_info[0].keys()
    local_edges = np.hstack([shared_info[0][key][1] for key in keys])
    global_edges = np.hstack([shared_info[1][key][1] for key in keys])

    own_edges = []
    for k, ie in enumerate(iedges):
        iii = np.where(local_edges == ie)[0]
        if len(iii) != 0:
            if ie == global_edges[iii[0]]:
               own_edges.append(ie)          
            iedges[k] = global_edges[iii[0]]
        else:
            own_edges.append(ie)

    #nicePrint("iedges", iedges)
    iedges_all = allgather_vector(iedges)
    dirs = allgather_vector(dirs)

    seens = defaultdict(int)
    seendirs  = defaultdict(int)        
    for k, ie in enumerate(iedges_all):
        seens[ie] = seens[ie] + 1
        seendirs[ie] = dirs[k]
    seens.default_factory = None

    idx = []
    signs = []
    for k in seens.keys():
        if seens[k] == 1:
            idx.append(k)
            signs.append(seendirs[k])
    iedges_g = np.array(idx)
    # here idx is global numbering
    
    #nicePrint("global_index", idx, signs)
    #nicePrint("own_edges", own_edges)
    
    iedges_l = []
    signs_l = []
    for k, ie in enumerate(iedges_g):
        iii = np.where(own_edges == ie)[0]
        if len(iii) != 0:
            iedges_l.append(ie)
            signs_l.append(signs[k])
    iedges_l = np.array(iedges_l) - offset_e[myid]
    signs_l = np.array(signs_l)

    #nicePrint("local_index", iedges_l, signs_l)

    return iedges_l, signs_l



import numpy as np

from petram.model import Model
from .solver_model import Solver
import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('StdSolver')
rprint = debug.regular_print('StdSolver')

class SolInit(Solver):
    can_delete = True
    has_2nd_panel = False
    mustbe_firstchild =True        

    def attribute_set(self, v):
        v['phys_model']   = ''
        super(SolInit, self).attribute_set(v)
        return v
    
    def panel1_param(self):
        return [["physics model",   self.phys_model,  0, {},],]

    def get_panel1_value(self):
        return (self.phys_model,)
    
    def import_panel1_value(self, v):
        self.phys_model = str(v[0])

    def get_editor_menus(self):
        return []

    def get_phys(self):
        names = self.phys_model.split(',')
        names = [n.strip() for n in names]
        return [self.root()['Phys'][n] for n in names]

    def init_sol(self, engine):
        phys_targets = self.get_phys()
        engine.run_init_sol(phys_targets)
        return

    def run(self, engine):
        phys_target = self.get_phys()
        self.init_sol(engine)




from petram.solver.solver_model import SolverInstance
import os
import numpy as np

from petram.model import Model
from petram.solver.solver_model import Solver
import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('StdSolver')
rprint = debug.regular_print('StdSolver')


class StdSolver(Solver):
    can_delete = True
    has_2nd_panel = False

    @classmethod
    def fancy_menu_name(self):
        return 'Stationary'

    @classmethod
    def fancy_tree_name(self):
        return 'Stationary'

    def verify_setting(self):
        try:
            phys = self.get_phys()
        except KeyError:
            return False, "", "Physics does not exist (KeyError)"
        if len(phys) == 0:
            return False, "", "No Physics is chosen to solve"

        return True, "", ""

    def attribute_set(self, v):
        super(StdSolver, self).attribute_set(v)
        return v

    def panel1_param(self):
        return [  # ["Initial value setting",   self.init_setting,  0, {},],
            ["physics model",   self.phys_model,  0, {}, ],
            [None, self.init_only,  3, {"text": "initialize solution only"}],
            [None,
             self.clear_wdir,  3, {"text": "clear working directory"}],
            [None,
             self.assemble_real,  3, {"text": "convert to real matrix (complex prob.)"}],
            [None,
             self.save_parmesh,  3, {"text": "save parallel mesh"}],
            [None,
             self.use_profiler,  3, {"text": "use profiler"}],
            [None, self.skip_solve,  3, {"text": "skip linear solve"}],
            [None, self.load_sol,  3, {
                "text": "load sol file (linear solver is not called)"}],
            [None, self.sol_file,  0, None], ]

    def get_panel1_value(self):
        return (  # self.init_setting,
            self.phys_model,
            self.init_only,
            self.clear_wdir,
            self.assemble_real,
            self.save_parmesh,
            self.use_profiler,
            self.skip_solve,
            self.load_sol,
            self.sol_file)

    def import_panel1_value(self, v):
        # self.init_setting = str(v[0])
        self.phys_model = str(v[0])
        self.init_only = v[1]
        self.clear_wdir = v[2]
        self.assemble_real = v[3]
        self.save_parmesh = v[4]
        self.use_profiler = v[5]
        self.skip_solve = v[6]
        self.load_sol = v[7]
        self.sol_file = v[8]

    def get_editor_menus(self):
        return []

    def get_possible_child(self):
        choice = []
        try:
            from petram.solver.mumps_model import MUMPS
            choice.append(MUMPS)
        except ImportError:
            pass

        # try:
        #    from petram.solver.gmres_model import GMRES
        #    choice.append(GMRES)
        # except ImportError:
        #    pass

        try:
            from petram.solver.iterative_model import Iterative
            choice.append(Iterative)
        except ImportError:
            pass

        try:
            from petram.solver.strumpack_model import Strumpack
            choice.append(Strumpack)
        except ImportError:
            pass
        return choice

    def allocate_solver_instance(self, engine):
        if self.clear_wdir:
            engine.remove_solfiles()

        instance = StandardSolver(self, engine)
        return instance

    def get_matrix_weight(self, timestep_config):  # , timestep_weight):
        if timestep_config[0]:
            return [1, 0, 0]
        else:
            return [0, 0, 0]

    @debug.use_profiler
    def run(self, engine, is_first=True, return_instance=False):
        dprint1("Entering run (is_first= ", is_first, ") ", self.fullpath())
        if self.clear_wdir:
            engine.remove_solfiles()

        instance = StandardSolver(
            self, engine) if self.instance is None else self.instance

        instance.set_blk_mask()
        if return_instance:
            return instance

        instance.configure_probes(self.probe)

        if self.init_only:
            engine.sol = engine.assembled_blocks[1][0]
            instance.sol = engine.sol

        elif self.load_sol:
            if is_first:
                instance.assemble()
                is_first = False
            instance.load_sol(self.sol_file)
        else:
            if is_first:
                instance.assemble()
                is_first = False
            else:
                instance.assemble(update=True)

            update_operator = engine.check_block_matrix_changed(
                instance.blk_mask)
            instance.solve(update_operator=update_operator)

        instance.save_solution(ksol=0,
                               skip_mesh=False,
                               mesh_only=False,
                               save_parmesh=self.save_parmesh)
        engine.sol = instance.sol

        instance.save_probe()

        self.instance = instance

        dprint1(debug.format_memory_usage())
        return is_first


class StandardSolver(SolverInstance):
    def __init__(self, gui, engine):
        SolverInstance.__init__(self, gui, engine)
        self.assembled = False
        self.linearsolver = None
        self._operator_set = False

    @property
    def blocks(self):
        return self.engine.assembled_blocks

    def compute_A(self, M, B, X, mask_M, mask_B):
        '''
        M[0] x = B

        return A and isAnew
        '''
        return M[0], np.any(mask_M[0])

    def compute_rhs(self, M, B, X):
        '''
        M[0] x = B
        '''
        return B

    def assemble(self, inplace=True, update=False):
        engine = self.engine
        phys_target = self.get_phys()
        phys_range = self.get_phys_range()

        # use get_phys to apply essential to all phys in solvestep
        dprint1("Asembling system matrix",
                [x.name() for x in phys_target],
                [x.name() for x in phys_range])

        if not update:
            engine.run_verify_setting(phys_target, self.gui)
        else:
            engine.set_update_flag('TimeDependent')

        M_updated = engine.run_assemble_mat(
            phys_target, phys_range, update=update)
        B_updated = engine.run_assemble_b(phys_target, update=update)

        engine.run_apply_essential(phys_target, phys_range, update=update)
        engine.run_fill_X_block(update=update)

        _blocks, M_changed = self.engine.run_assemble_blocks(self.compute_A,
                                                             self.compute_rhs,
                                                             inplace=inplace,
                                                             update=update,)
        # A, X, RHS, Ae, B, M, names = blocks
        self.assembled = True
        return M_changed

    def assemble_rhs(self):
        engine = self.engine
        phys_target = self.get_phys()
        engine.run_assemble_b(phys_target)
        B = self.engine.run_update_B_blocks()
        self.blocks[4] = B
        self.assembled = True

    def solve(self, update_operator=True):
        update_operator = update_operator or not self._operator_set
        engine = self.engine

        # if not self.assembled:
        #    assert False, "assmeble must have been called"

        A, X, RHS, Ae, B, M, depvars = self.blocks
        mask = self.blk_mask
        engine.copy_block_mask(mask)

        depvars = [x for i, x in enumerate(depvars) if mask[0][i]]

        if update_operator:
            AA = engine.finalize_matrix(A, mask, not self.phys_real,
                                        format=self.ls_type)

        BB = engine.finalize_rhs([RHS], A, X[0], mask, not self.phys_real,
                                 format=self.ls_type)

        if self.linearsolver is None:
            linearsolver = self.allocate_linearsolver(
                self.gui.is_complex(), self. engine)
            self.linearsolver = linearsolver
        else:
            linearsolver = self.linearsolver

        linearsolver.skip_solve = self.gui.skip_solve

        if update_operator:
            linearsolver.SetOperator(AA,
                                     dist=engine.is_matrix_distributed,
                                     name=depvars)
            self._operator_set = True

        if linearsolver.is_iterative:
            XX = engine.finalize_x(X[0], RHS, mask, not self.phys_real,
                                   format=self.ls_type)
        else:
            XX = None

        solall = linearsolver.Mult(BB, x=XX, case_base=0)

        from petram.mfem_config import use_parallel
        if use_parallel:
            from mpi4py import MPI
        else:
            from petram.helper.dummy_mpi import MPI
        myid = MPI.COMM_WORLD.rank
        size = MPI.COMM_WORLD.size

        ss = str(solall.shape) if solall is not None else "x"
        ss = MPI.COMM_WORLD.gather(ss)
        if ss is not None:
            dprint1("solshape", ', '.join(ss))

        # linearsolver.SetOperator(AA, dist = engine.is_matrix_distributed)
        # solall = linearsolver.Mult(BB, case_base=0)

        is_sol_central = any(MPI.COMM_WORLD.allgather(solall is None))
        if is_sol_central:
            if not self.phys_real and self.gui.assemble_real:
                solall = self.linearsolver_model.real_to_complex(solall, AA)
            A.reformat_central_mat(solall, 0, X[0], mask)
        else:
            if not self.phys_real and self.gui.assemble_real:
                assert False, "this operation is not permitted"
            A.reformat_distributed_mat(solall, 0, X[0], mask)

        self.sol = X[0]

        # store probe signal (use t=0.0 in std_solver)
        for p in self.probe:
            p.append_sol(X[0])

        return True

    def load_sol(self, solfile):
        from petram.mfem_config import use_parallel
        if use_parallel:
            from mpi4py import MPI
        else:
            from petram.helper.dummy_mpi import MPI
        myid = MPI.COMM_WORLD.rank

        if myid == 0:
            solall = np.load(solfile)
        else:
            solall = None

        A, X, RHS, Ae, B, M, depvars = self.blocks
        mask = self.blk_mask
        A.reformat_central_mat(solall, 0, X[0], mask)
        self.sol = X[0]

        # store probe signal (use t=0.0 in std_solver)
        for p in self.probe:
            p.append_sol(X[0])

        return True



from petram.helper.matrix_file import write_coo_matrix, write_vector
import sys
import os
import numpy as np
import scipy.sparse
from collections import OrderedDict
from warnings import warn

from petram.mfem_config import use_parallel
if use_parallel:
    from petram.helper.mpi_recipes import *
    import mfem.par as mfem
else:
    import mfem.ser as mfem
import mfem.common.chypre as chypre

# these are only for debuging
from mfem.common.parcsr_extra import ToScipyCoo
from mfem.common.mpi_debug import nicePrint

from petram.phys.phys_model import Phys
from petram.model import Domain, Bdry, ModelDict
from petram.phys.vtable import VtableElement, Vtable


import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('AUX_Operator')

#groups = ['Domain', 'Boundary', 'Edge', 'Point', 'Pair']
groups = ['Domain', 'Boundary', 'Pair']

data = [("coeff_lambda", VtableElement("coeff_lambda", type='array',
                                       guilabel="lambda",
                                       default=1.0,
                                       tip="coefficient",)),
        ("axu_oprt", VtableElement("aux_oprt", type='any',
                                   guilabel="operator",
                                   default="",
                                   no_func=True,
                                   tip="oprator (horizontal)",)), ]


class AUX_Operator(Phys):
    vt_oprt = Vtable(data)
    has_3rd_panel = True
    _has_4th_panel = True

    def attribute_set(self, v):
        v = super(AUX_Operator, self).attribute_set(v)
        v['paired_var'] = None  # (phys_name, index)
        v['src_var'] = 0  # (index)
        v['use_symmetric'] = False
        v['use_anti_symmetric'] = False
        v['no_elimination'] = False
        v['use_conj'] = False
        v['coeff_type'] = 'Scalar'
        v = self.vt_oprt.attribute_set(v)
        #vv = self.vt_oprt.attribute_set({})
        # for key in vv:
        #    if hasattr(self, key): vv[key] = getattr(self, key)
        #    v[key] = vv[key]
        return v

    def save_attribute_set(self, skip_def_check):
        attrs = super(AUX_Operator, self).save_attribute_set(skip_def_check)
        vv = self.vt_oprt.attribute_set({})
        for key in vv:
            if not key in attrs:
                attrs.append(key)
        return attrs

    def panel1_param(self):
        import wx

        mfem_physroot = self.get_root_phys().parent
        names, pnames, pindex = mfem_physroot.dependent_values()
        names = [n+" ("+p + ")" for n, p in zip(names, pnames)]

        dep_vars = self.get_root_phys().dep_vars

        ll1 = [["trial space (Cols)", names[0], 4,
                {"style": wx.CB_READONLY, "choices": names}],
               ["test space (Rows)", dep_vars[0], 4,
                {"style": wx.CB_READONLY, "choices": dep_vars}],
               ["coeff. type", "Scalar", 4,
                {"style": wx.CB_READONLY,
                 "choices": ["Scalar", "Vector", "Diagonal", "Matrix"]}, ], ]
        ll2 = self.vt_oprt.panel_param(self)
        ll3 = [["symmetric",  self.use_symmetric,        3, {"text": ""}],
               ["anti-symmetric",  self.use_anti_symmetric,   3, {"text": ""}],
               ["use  conjugate",  self.use_conj,             3, {"text": ""}],
               ["use as ess. Dof",  self.no_elimination,   3, {"text": ""}], ]

        return ll1 + ll2 + ll3

    def import_panel1_value(self, v):
        mfem_physroot = self.get_root_phys().parent
        names, pnames, pindex = mfem_physroot.dependent_values()

        idx = names.index(str(v[0]).split("(")[0].strip())
        self.paired_var = (pnames[idx], pindex[idx])

        self.src_var = self.get_root_phys().dep_vars.index(str(v[1]))
        self.coeff_type = str(v[2])
        self.vt_oprt.import_panel_value(self, v[3:-4])
        self.use_symmetric = v[-4]
        self.use_anti_symmetric = v[-3]
        self.use_conj = v[-2]
        self.no_elimination = v[-1]

    def get_panel1_value(self):
        if self.paired_var is None:
            n = self.get_root_phys().dep_vars[0]
            p = self.get_root_phys().name()
        else:
            mfem_physroot = self.get_root_phys().parent
            if self.paired_var[0] in mfem_physroot:
                var_s = mfem_physroot[self.paired_var[0]].dep_vars
                n = var_s[self.paired_var[1]]
                p = self.paired_var[0]
            else:
                # if paird var does not exist in the model
                n = self.get_root_phys().dep_vars[0]
                p = self.get_root_phys().name()

        var = n + " ("+p + ")"

        svar = self.get_root_phys().dep_vars[self.src_var]

        v1 = [var, svar, self.coeff_type]
        val, expr = self.vt_oprt.get_panel_value(self)
        if expr.strip().startswith('='):
            expr = expr.strip()[1:]
        v1.extend([val, expr])
        v3 = [self.use_symmetric, self.use_anti_symmetric,
              self.use_conj, self.no_elimination]
        return v1 + v3

    def panel2_param(self):
        return [[None, "Auxiriary varialbe is global",  2,   {}], ]

    def import_panel2_value(self, v):
        pass

    def get_panel2_value(self):
        return [None]

    def has_extra_DoF(self, kfes):
        return False

    def get_extra_NDoF(self):
        return 0

    def verify_setting(self):
        policy = self.root()['General'].diagpolicy
        if policy == 'keep' and self.no_elimination:
            return (False, "DiagKeep with AUX Ebssential",
                    "Diag Policy is not one but used as essential")
        else:
            return True, "", ""

    def preprocess_params(self, engine):
        self.vt_oprt.preprocess_params(self)
        super(AUX_Operator, self).preprocess_params(engine)

    def has_aux_op(self, phys1, kfes, phys2, kfes2):
        # check
        trialname2 = phys2.dep_vars[kfes2]
        testname2 = phys1.dep_vars[kfes]

        mfem_physroot = self.get_root_phys().parent
        var_s = mfem_physroot[self.paired_var[0]].dep_vars
        trialname = var_s[self.paired_var[1]]
        testname = self.get_root_phys().dep_vars[self.src_var]

        if (trialname == trialname2 and
                testname == testname2):
            return True

        if (trialname == testname2 and
            testname == trialname2 and
                (self.use_symmetric or self.use_anti_symmetric)):
            return True

        return False

    def get_aux_op(self, engine, phys1, kfes1, phys2, kfes2,
                   trial_ess_tdof=None,
                   test_ess_tdof=None):

        mfem_physroot = self.get_root_phys().parent

        var_s = mfem_physroot[self.paired_var[0]].dep_vars
        trialname = var_s[self.paired_var[1]]
        testname = self.get_root_phys().dep_vars[self.src_var]

        oprt = self.aux_oprt_txt.strip()
        if oprt.startswith('='):  # old setting
            oprt = oprt[1:]
        coeff = self.vt_oprt['coeff_lambda'].make_value_or_expression(self)

        from petram.helper.expression import Expression

        fes1 = engine.fespaces[trialname]
        fes2 = engine.fespaces[testname]
        ind_vars = self.get_root_phys().ind_vars
        is_complex = self.get_root_phys().is_complex()

        diag_size = -1

        if oprt is not None:
            dprint1(self.name() + " Assembling Operator: ", oprt)
            assert isinstance(oprt, str), "operator1 must be an expression"

            cotype = self.coeff_type[0]
            c_coeff1 = self.get_coefficient_from_expression(coeff, cotype,
                                                            use_dual=False,
                                                            real=True,
                                                            is_conj=False)
            if is_complex:
                c_coeff2 = self.get_coefficient_from_expression(coeff, cotype,
                                                                use_dual=False,
                                                                real=False,
                                                                is_conj=False)
            else:
                c_coeff2 = None
            c_coeff = (c_coeff1, c_coeff2)

            expr = Expression(oprt, engine=engine, trial=fes1, test=fes2,
                              trial_ess_tdof=trial_ess_tdof,
                              test_ess_tdof=test_ess_tdof,
                              ind_vars=ind_vars,
                              is_complex=is_complex,
                              c_coeff=c_coeff)
            op = expr.assemble(g=self._global_ns)

        trialname2 = phys2.dep_vars[kfes2]

        if testname == trialname2 and (self.use_symmetric or self.use_anti_symmetric):
            op = op.transpose()
            if self.use_anti_symmetric:
                op *= -1.0
            if self.use_conj:
                op = op.conj()

        return op



import wx
import numpy as np
from collections import defaultdict

import petram
from petram.utils import get_pkg_datafile

fdotbk = get_pkg_datafile(petram.pi, 'icon',  'dot_bk.png')
fedgebk = get_pkg_datafile(petram.pi, 'icon', 'line_bk.png')
ffacebk = get_pkg_datafile(petram.pi, 'icon', 'face_bk.png')
fdot = get_pkg_datafile(petram.pi, 'icon',  'dot.png')
fedge = get_pkg_datafile(petram.pi, 'icon', 'line.png')
fface = get_pkg_datafile(petram.pi, 'icon', 'face.png')
fdom = get_pkg_datafile(petram.pi, 'icon', 'domain.png')
showall = get_pkg_datafile(petram.pi, 'icon', 'showall.png')
fshow = get_pkg_datafile(petram.pi, 'icon', 'show.png')
hide = get_pkg_datafile(petram.pi, 'icon', 'hide.png')
fsolid = get_pkg_datafile(petram.pi, 'icon', 'solid.png')
ftrans = get_pkg_datafile(petram.pi, 'icon', 'transparent.png')


from petram.pi.sel_buttons import _select_x

def select_dot(evt):
    _select_x(evt, 'point', 'point')
    
def select_edge(evt):
    _select_x(evt, 'edge', 'edge')
    
def select_face(evt):
    _select_x(evt, 'face', 'face')
    
def select_volume(evt):
    _select_x(evt, 'volume', 'face')
    
def show_all(evt):
    viewer = evt.GetEventObject().GetTopLevelParent()
    mode = viewer._sel_mode

    ax = viewer.get_axes()
    
    Fobjs = [child for name, child in ax.get_children() if name.startswith("face")
            and not name.endswith("meshed")]
    mFobjs = [child for name, child in ax.get_children() if name.startswith("face")
            and name.endswith("meshed")]
    Eobjs = [child for name, child in ax.get_children() if name.startswith("edge")
            and not name.endswith("meshed")]
    mEobjs = [child for name, child in ax.get_children() if name.startswith("edge")
            and name.endswith("meshed")]


    def show_all_obj(objs, meshed_objs):
        idx = []        
        for o in meshed_objs:
            if o.hasvar('idxset'):
                o.hide_component([])
            idx = idx + list(np.unique(o.getvar('array_idx')))
        for o in objs:
            if o.hasvar('idxset'):            
                o.hide_component(idx)        
    
    if mode == 'volume':
         show_all_obj(Fobjs, mFobjs)
         show_all_obj(Eobjs, mEobjs)         
    elif mode == 'face':
         show_all_obj(Fobjs, mFobjs)
         show_all_obj(Eobjs, mEobjs)         
    elif mode == 'edge':
         show_all_obj(Fobjs, mFobjs)
         show_all_obj(Eobjs, mEobjs)         
    elif mode == 'point':
        if ax.point.hasvar('idxset'):            
             ax.point.hide_component([])                        
    else:
        pass
    
    viewer._mhidden_volume = []
    viewer._mhidden_face = []
    viewer._mhidden_edge = []
    viewer._dom_bdr_sel  = ([], [], [], [])            
    
    viewer.draw_all()    

def hide_elem(evt, inverse=False):
    viewer = evt.GetEventObject().GetTopLevelParent()
    mode = viewer._sel_mode
    ax = viewer.get_axes()
    
    namestart = mode if mode != 'volume' else 'face'
    namestart2 =  namestart + '_meshed'    
    objs = [child for name, child in ax.get_children() if name.startswith(namestart)
            and not name.startswith(namestart2)]
    meshed_objs = [child for name, child in ax.get_children() if name.startswith(namestart2)]
    
    Eobjs = [child for name, child in ax.get_children() if name.startswith('edge')]
    Fobjs = [child for name, child in ax.get_children() if name.startswith('face')]

    sel = viewer.canvas.selection
    #if len(sel) != 1:  return
    sel_objs = [s().figobj for s in sel]

    s2l, v2s = viewer._s_v_loop['mesh']    
    if mode == 'volume':
        facesa = []
        facesb = []        
        
        selected_volume = viewer._dom_bdr_sel[0]
        target_volumes = selected_volume

        if not inverse:
            target_volumes.extend(viewer._mhidden_volume)
        for key in v2s.keys():
            if key in target_volumes:
                facesa.extend(v2s[key])
            else:
                facesb.extend(v2s[key])
        if inverse:
            for o in objs:                         
                o.hide_component(facesa, inverse=True)
            for o in meshed_objs:                         
                o.hide_component(facesa, inverse=True)
            hidden_volume = [x for x in v2s.keys() if not x
                             in selected_volume]            
            viewer._mhidden_volume = hidden_volume
        else:
            facesa = np.unique(np.array(facesa))
            facesb = np.unique(np.array(facesb))
            new_hide = list(np.setdiff1d(facesa, facesb, True))
            for o in objs:                                     
                idx = o.hidden_component
                idx = list(set(idx+new_hide))
                o.hide_component(idx)
            for o in meshed_objs:                                     
                idx = o.hidden_component
                idx = list(set(idx+new_hide))
                o.hide_component(idx)
            tmp = viewer._mhidden_volume
            tmp.extend(selected_volume)
            viewer._mhidden_volume = list(set(tmp))
    elif mode == 'face' or mode == 'edge':
        idx1 = []
        for o in objs + meshed_objs:
            idx1.extend(o.getSelectedIndex())
        for o in objs + meshed_objs:
            idx = list(set(o.hidden_component+idx1))        
            o.hide_component(idx, inverse=inverse)
        if mode == "face":
            if inverse:
                hidden_face = [x for x in s2l.keys() if not x
                                 in idx]
            else:
                hidden_face = idx
            viewer._mhidden_face = hidden_face
        if mode == "edge":
            if inverse:
                all_edges = list(np.unique(np.hstack(([s2l[x] for x in s2l.keys()]))))
                hidden_edge = [x for x in all_edges if not x in idx]
            else:
                hidden_edge = idx
            viewer._mhidden_edge = hidden_edge
    elif mode == 'point':
        pass
    else:
        pass

    if mode == 'volume' or mode == 'face':
        from petram.mesh.mesh_utils import line2surf        
        hidden_face = sum([o.hidden_component for o in Fobjs], [])
        l2s = line2surf(s2l)
        
        dd = defaultdict(int)
        for f in hidden_face:
            for x in s2l[f]:
                dd[x] = dd[x]+1
        dd = dict(dd)
        hide_this_edge = [x for x in dd if dd[x] == len(l2s[x])]
        
        #hide_this_edge = [l for l, ss in l2s.items()
        #              if np.intersect1d(ss, hidden_face).size==len(ss)]

        for o in Eobjs:
            #idx = o.getSelectedIndex()
            idx = list(set(o.hidden_component+hide_this_edge))
            if o.hasvar('idxset'):            
                o.hide_component(idx)
    
    viewer.canvas.unselect_all()    
    viewer.draw_all()
    
def show_only(evt):    
    hide_elem(evt, inverse=True)

def _toggle_any(evt, txt):
    viewer = evt.GetEventObject().GetTopLevelParent()
    mode = viewer._sel_mode

    ax = viewer.get_axes()
    objs = [child for name, child in ax.get_children() if name.startswith(txt)]

    c_status = any([o.isSuppressed for o in objs])
    for o in objs:
        if c_status:
            o.onUnSuppress()
        else:
            o.onSuppress()
        wx.CallAfter(o.set_gl_hl_use_array_idx, True)
    viewer.canvas.unselect_all()
    evt.Skip()
    
def toggle_dot(evt):
    _toggle_any(evt, 'point')
def toggle_edge(evt):
    _toggle_any(evt, 'edge')
    
from petram.pi.sel_buttons import toggle_face    

def make_solid(evt):
    viewer = evt.GetEventObject().GetTopLevelParent()
    mode = viewer._sel_mode

    ax = viewer.get_axes()
    for name, child in ax.get_children():
        if len(child._artists) > 0:
            child.set_alpha(1.0, child._artists[0])
    viewer.canvas.unselect_all()
    viewer.draw_all()
    
def make_transp(evt):
    viewer = evt.GetEventObject().GetTopLevelParent()
    mode = viewer._sel_mode

    ax = viewer.get_axes()
    for name, child in ax.get_children():
        if len(child._artists) > 0:
            child.set_alpha(0.75, child._artists[0])
    viewer.canvas.unselect_all()
    viewer.draw_all()
    
btask = [('mdot',    fdot,  2, 'select vertex', select_dot),
         ('medge',   fedge, 2, 'select edge', select_edge),
         ('mface',   fface, 2, 'select face', select_face),
         ('mdomain', fdom,  2, 'select domain', select_volume),
         ('---', None, None, None),
         ('mtoggledot',    fdotbk,  0, 'toggle vertex', toggle_dot),
         ('mtoggleedge',   fedgebk, 0, 'toggle edge', toggle_edge),
         ('mtoggleface',   ffacebk, 0, 'toggle face', toggle_face),             
         ('---', None, None, None),
         ('mshow',   showall,  0, 'show all', show_all),
         ('mhide',   hide,  0, 'hide selection', hide_elem),        
         ('mshowonly',  fshow,  0, 'show only', show_only),    
         ('---', None, None, None),
         ('msolid',  fsolid,  0, 'solid', make_solid),
         ('mtranspaent',  ftrans,  0, 'transparent', make_transp),]
            



import os
import traceback
import gc

from petram.model import Model
from petram.solver.solver_model import Solver, SolveStep
from petram.namespace_mixin import NS_mixin
import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('Parametric')
format_memory_usage = debug.format_memory_usage

assembly_methods = {'Full assemble': 0,
                    'Reuse matrix' : 1}

def def_start_cond(count, max_count,  *args, **kwargs):
    return False

def def_stop_cond(count, max_count, *args, **kwargs):
    return count >= max_count
    
class Loop(SolveStep, NS_mixin):
    has_2nd_panel = False
    
    def __init__(self, *args, **kwargs):
        SolveStep.__init__(self, *args, **kwargs)
        NS_mixin.__init__(self, *args, **kwargs)
    
    def attribute_set(self, v):
        v['phys_model']   = ''
        v['init_setting']   = ''
        v['postprocess_sol']   = ''        
        v['use_dwc_pp']   = False
        v['dwc_pp_arg']   = ''

        v['max_count'] = 1
        v['start_cond_func'] = ''        
        v['stop_cond_func'] = ''
        super(Loop, self).attribute_set(v)
        return v
    
    def panel1_param(self):

        ret = super(Loop, self).panel1_param()
        ret = ret +[["Max loop count'",    self.max_count,   400, {},],
                    ["Start cond.",    self.start_cond_func,  0, {},],        
                    ["Stop cond.",     self.stop_cond_func,   0, {},],]
                   
        return ret

    def get_panel1_value(self):
        ret = list(super(Loop, self).get_panel1_value())

        ret = ret + [self.max_count,
                     self.start_cond_func,
                     self.stop_cond_func]
        
        return ret

    def import_panel1_value(self, v):
        super(Loop, self).import_panel1_value(v[:-3])
        self.max_count = int(v[-3])
        self.start_cond_func = v[-2]
        self.stop_cond_func = v[-1]

    def run(self, engine, is_first = True):
        dprint1("Entering SolveStep :" + self.name())
        solvers = self.get_active_solvers()

        # initialize and assemble here

        # in run method..
        #   std solver : make sub block matrix and solve
        #   time-domain solver : do step

        start_func = def_start_cond
        stop_func = def_stop_cond

        if self.start_cond_func in self._global_ns:
             start_func = self._global_ns[self.start_cond_func]
        if self.stop_cond_func in self._global_ns:
             stop_func = self._global_ns[self.stop_cond_func]
                                         
        for i in range(self.max_count):
            self.prepare_form_sol_variables(engine)
            self.init(engine)

            is_first = True            
            if start_func(i, self.max_count): break
            for solver in solvers:
                 is_first = solver.run(engine, is_first=is_first)
                 engine.add_FESvariable_to_NS(self.get_phys()) 
                 engine.store_x()
                 if self.solve_error[0]:
                     dprint1("Loop failed " + self.name() + ":"  + self.solve_error[1])
                     break
            if stop_func(i, self.max_count): break
             
        postprocess = self.get_pp_setting()
        engine.run_postprocess(postprocess, name = self.name())
        
        if self.use_dwc_pp:
            engine.call_dwc(self.get_phys_range(),
                            method="postprocess",
                            callername = self.name(),
                            args = self.dwc_pp_arg)
    



#
#  petra
#
#   piScope command to start PetraM
#
import os


class PetraMHelper(object):
    '''
    configurator/helper
    '''

    def __init__(self, p=None):
        object.__init__(self)
        self.properties = {'refine': 5}

        if p is not None:
            for k in p:
                if k in self.properties:
                    self.properties[k] = p[k]
        for k in self.properties:
            self.set(k, self.properties[k])

    def set(self, name, value):
        if name == 'refine':
            import petram.mesh.refined_mfem_geom
            petram.mesh.refined_mfem_geom.default_refine = value

    def set_debug(self, name, value=True):
        import petram.debug
        assert hasattr(petram.debug, "debug_"+name), "debug_" + \
            name+" is not valid debug option."
        setattr(petram.debug, "debug_"+name, value)

    def show(self):
        for k in self.properties:
            print(k + ' : ' + str(self.properties[k]))

    def versions(self):
        import mfem
        import wx
        import matplotlib
        import OCC

        print("WX:", wx.__version__)
        print("Matplotlib:", matplotlib.__version__)
        print("OCC:", OCC.VERSION)
        print("MFEM:", mfem.__version__)


def petram(reload_scripts=False, addon="none", cluster=False):
    '''
    setup PetraM simulation enveroment

       cluster : turn-on cluster access menu, temporariliy.
       reload_scripts : reload project sciripts
       addon : turn-on addon access, temporariliy.
    '''
    import petram.mfem_model as mm
    if not mm.has_cluster_access and cluster is True:
        mm.has_cluster_access = True
    if mm.has_addon_access == "none":
        mm.has_addon_access = addon

    import wx
    ifig_app = wx.GetApp().TopWindow
    proj = ifig_app.proj
    if proj.setting.parameters.hasvar('PetraM'):
        model = proj.setting.parameters.eval('PetraM')
    else:
        try:
            model = proj.model1.mfem
            scripts = model.scripts
            from ifigure.mto.hg_support import has_repo
            if has_repo(scripts):
                scripts.onHGturnoff(evt=None, confirm=False)
                model.param.setvar('remote', None)
            reload_scripts = True
        except:
            model = load_petra_model(proj)
    if reload_scripts:
        scripts = model.scripts
        for name, child in scripts.get_children():
            child.destroy()
        scripts.clean_owndir()
        import_project_scripts(scripts)

    if model is not None:
        if not model.has_child('mfembook'):
            book = model.add_book('mfembook')
            ipage = book.add_page()
            book.get_page(ipage).add_axes()
            book.set_keep_data_in_tree(True)

        from petram.mfem_viewer import MFEMViewer
        viewer = model.mfembook.Open(MFEMViewer)
        viewer.isec(0)
        viewer.threed('on')
        viewer.view('noclip')
        proj.setting.parameters.setvar('PetraM', '='+model.get_full_path())

        import wx
        wx.CallAfter(model.mfembook.find_bookviewer().Raise)

    return PetraMHelper()


def load_petra_model(proj):

    model_root = proj.onAddModel()
    model = model_root.add_model('mfem')
    model.onAddNewNamespace(e=None)

    model.param.setvar('nproc', 2)
    model.param.setvar('openmp_num_threads', 'auto')
    model.param.setvar('openblas_num_threads', 1)
    model.add_folder('namespaces')
    model.add_folder('datasets')
    model.add_folder('solutions')
    scripts = model.add_folder('scripts')
    import_project_scripts(scripts)

    scripts.helpers.reset_model()
    # model.set_guiscript('.scripts.helpers.open_gui')
    model.scripts.helpers.create_ns('global')

    param = model.param
    param.setvar('mfem', None)
    param.setvar('sol', None)
    param.setvar('remote', None)

    return model


def import_project_scripts(scripts):
    import petram.pi.project_scripts

    path = os.path.dirname(petram.pi.project_scripts.__file__)
    scripts.load_script_folder(path, skip_underscore=True)
